#!/usr/bin/env python

# Copyright 2007 The Spitfire Authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.


from future import standard_library
import importlib
standard_library.install_aliases()
from builtins import str
from builtins import object
import copy
import logging
import optparse
import os.path
import sys
import time
import traceback

import io as StringIO

from spitfire.compiler import compiler
from spitfire.compiler import options
from spitfire.compiler import util
from spitfire.compiler import visitor
from spitfire import runtime
from spitfire.runtime import runner
from spitfire.runtime import udn


# this class let's me check if placeholder caching is working properly by
# tracking the number of accesses for a single key
class ResolveCounter(object):

    def __init__(self):
        self._dict = {}

    @property
    def resolve_x(self):
        return self._get_item('resolve_x')

    @property
    def resolve_y(self):
        return self._get_item('resolve_y')

    def _get_item(self, key):
        if key in self._dict:
            self._dict[key] += 1
        else:
            self._dict[key] = 1
        return '%s%s' % (key, self._dict[key])

    def __contains__(self, key):
        return key.startswith('resolve')

    def __getitem__(self, key):
        if not key.startswith('resolve'):
            raise KeyError(key)
        return self._get_item(key)

    def __getattr__(self, key):
        if not key.startswith('resolve'):
            raise AttributeError(key)
        return self._get_item(key)


sys_modules = list(sys.modules.keys())


def reset_sys_modules():
    for key in list(sys.modules.keys()):
        if key not in sys_modules:
            del sys.modules[key]


class TestRunner(object):

    def __init__(self, spt_compiler, spt_options, spt_files):
        self.compiler = spt_compiler
        self.options = spt_options
        self.files = spt_files
        self._search_list = [
            {'tier1': {'tier2': ResolveCounter()}},
            {'nest': ResolveCounter()},
            ResolveCounter(),
        ]
        if self.options.test_input:
            self._search_list.append(runner.load_search_list(
                self.options.test_input))
        self.buffer = StringIO.StringIO()
        self.start_time = 0
        self.finish_time = 0
        self.num_tests_run = 0
        self.num_tests_failed = 0

    # return a copy of the search_list for each set of tests
    @property
    def search_list(self):
        return copy.deepcopy(self._search_list)

    def run(self):
        self.begin()
        for filename in self.files:
            self.process_file(filename)
        self.end()

    def begin(self):
        self.start_time = time.time()

    def end(self):
        self.finish_time = time.time()
        print(file=sys.stderr)
        if self.num_tests_failed > 0:
            sys.stderr.write(self.buffer.getvalue())
        print('-' * 70, file=sys.stderr)
        print('Ran %d tests in %0.3fs' % (
            self.num_tests_run, self.finish_time - self.start_time), file=sys.stderr)
        print(file=sys.stderr)
        if self.num_tests_failed > 0:
            print('FAILED (failures=%d)' % self.num_tests_failed, file=sys.stderr)
            sys.exit(1)
        else:
            print('OK', file=sys.stderr)
            sys.exit(0)

    def process_file(self, filename):
        buffer = StringIO.StringIO()
        reset_sys_modules()

        classname = util.filename2classname(filename)
        modulename = util.filename2modulename(filename)
        test_output_path = os.path.join(self.options.test_output,
                                        classname + '.txt')

        if self.options.verbose:
            sys.stderr.write(modulename + ' ... ')

        compile_failed = False
        if self.options.debug or self.options.compile:
            try:
                self.compiler.compile_file(filename)
            except Exception as e:
                compile_failed = True
                print('=' * 70, file=buffer)
                print('FAIL:', modulename, '(' + filename + ')', file=buffer)
                print('-' * 70, file=buffer)
                traceback.print_exc(None, buffer)
            if self.options.debug:
                if 'parse_tree' in self.options.debug_flags:
                    print("parse_tree:", file=buffer)
                    visitor.print_tree(self.compiler._parse_tree, output=buffer)
                if 'analyzed_tree' in self.options.debug_flags:
                    print("analyzed_tree:", file=buffer)
                    visitor.print_tree(self.compiler._analyzed_tree,
                                       output=buffer)
                if 'optimized_tree' in self.options.debug_flags:
                    print("optimized_tree:", file=buffer)
                    visitor.print_tree(self.compiler._optimized_tree,
                                       output=buffer)
                if 'hoisted_tree' in self.options.debug_flags:
                    print("hoisted_tree:", file=buffer)
                    visitor.print_tree(self.compiler._hoisted_tree,
                                       output=buffer)
                if 'source_code' in self.options.debug_flags:
                    print("source_code:", file=buffer)
                    for i, line in enumerate(self.compiler._source_code.split(
                            '\n')):
                        print('% 3s' % (i + 1), line, file=buffer)

        test_failed = False
        if not self.options.skip_test:
            import tests

            current_output = None
            raised_exception = False
            try:
                if self.options.debug or self.options.compile:
                    template_module = util.load_module_from_src(
                        self.compiler._source_code, filename, modulename)
                else:
                    template_module = runtime.import_module_symbol(modulename)
            except Exception as e:
                # An exception here means the template is unavailble; the test
                # fails.
                test_failed = True
                raised_exception = True
                current_output = str(e)

            if not test_failed:
                try:
                    template_class = getattr(template_module, classname)
                    template = template_class(search_list=self.search_list)
                    current_output = template.main().encode('utf8')
                except Exception as e:
                    # An exception here doesn't meant that the test fails
                    # necessarily since libraries don't have a class; as long as
                    # the expected output matches the exception, the test
                    # passes.
                    raised_exception = True
                    current_output = str(e)

            if not test_failed:
                if self.options.test_accept_result:
                    test_file = open(test_output_path, 'w')
                    test_file.write(current_output)
                    test_file.close()
                try:
                    test_file = open(test_output_path)
                except IOError as e:
                    # An excpetion here means that the expected output is
                    # unavailbe; the test fails.
                    test_failed = True
                    raised_exception = True
                    current_output = str(e)

            if test_failed:
                test_output = None
            else:
                test_output = test_file.read()
                if current_output != test_output:
                    test_failed = True
                    if self.options.debug:
                        print("expected output:", file=buffer)
                        print(test_output, file=buffer)
                        print("actual output:", file=buffer)
                        print(current_output, file=buffer)

            if compile_failed or test_failed:
                self.num_tests_failed += 1
                if self.options.verbose:
                    sys.stderr.write('FAIL\n')
                else:
                    sys.stderr.write('F')
                current_output_path = os.path.join(self.options.test_output,
                                                   classname + '.failed')
                f = open(current_output_path, 'w')
                f.write(current_output)
                f.close()
                print('=' * 70, file=buffer)
                print('FAIL:', modulename, '(' + filename + ')', file=buffer)
                print('-' * 70, file=buffer)
                print('Compare expected and actual output with:', file=buffer)
                print(' '.join(['    diff -u', test_output_path,
                                           current_output_path]), file=buffer)
                print('Show debug information for the test with:', file=buffer)
                test_cmd = [arg for arg in sys.argv if arg not in self.files]
                if '--debug' not in test_cmd:
                    test_cmd.append('--debug')
                test_cmd = ' '.join(test_cmd)
                print('   ', test_cmd, filename, file=buffer)
                if raised_exception:
                    print('-' * 70, file=buffer)
                    print(current_output, file=buffer)
                    traceback.print_exc(None, buffer)
                print(file=buffer)
                self.buffer.write(buffer.getvalue())
            else:
                if self.options.verbose:
                    sys.stderr.write('ok\n')
                else:
                    sys.stderr.write('.')
            self.num_tests_run += 1


if __name__ == '__main__':
    importlib.reload(sys)
    sys.setdefaultencoding('utf8')

    option_parser = optparse.OptionParser()
    options.add_common_options(option_parser)
    option_parser.add_option('-c',
                             '--compile',
                             action='store_true',
                             default=False)
    option_parser.add_option('--skip-test', action='store_true', default=False)
    option_parser.add_option(
        '--test-input',
        default='tests/input/search_list_data.pye',
        help='input data file for templates (.pkl or eval-able file)')
    option_parser.add_option('--test-output',
                             default='tests/output',
                             help="directory for output")
    option_parser.add_option(
        '--test-accept-result',
        action='store_true',
        default=False,
        help='accept current code output as correct for future tests')
    option_parser.add_option('--debug', action='store_true', default=False)
    option_parser.add_option(
        '--debug-flags',
        action='store',
        default='hoisted_tree,source_code',
        help='parse_tree, analyzed_tree, optimized_tree, hoisted_tree, source_code')
    option_parser.add_option('--enable-c-accelerator',
                             action='store_true',
                             default=False)

    (spt_options, spt_args) = option_parser.parse_args()
    if spt_options.debug:
        spt_options.verbose = True
        spt_options.debug_flags = getattr(spt_options, 'debug_flags').split(',')
    else:
        spt_options.debug_flags = []

    udn.set_accelerator(spt_options.enable_c_accelerator, enable_test_mode=True)

    spt_compiler_args = compiler.Compiler.args_from_optparse(spt_options)
    spt_compiler = compiler.Compiler(**spt_compiler_args)

    test_runner = TestRunner(spt_compiler, spt_options, spt_args)
    test_runner.run()



# Copyright 2007 The Spitfire Authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import builtins
import copy
import traceback


# this is a horrible hack to let the tree modify itself during conversion
class EatPrevious(object):
    pass


class ASTNode(object):

    def __init__(self, name='', pos=None):
        self.name = name
        self.value = None
        self.parent = None
        self.child_nodes = NodeList()
        # optimization annotations
        self.hint_map = {}
        # Start and end position of a text or placeholder node within an I18n'd
        # message.
        self.start = self.end = None
        # tag items generated from a single line statment
        # this makes it easier to invent new ways to mangle optional whitespace
        self.statement = False
        # Position in the input string (measured in characters).
        self.pos = pos
        self.end_pos = None

    def __str__(self):
        if self.value:
            return '%s %s %r' % (self.__class__.__name__, self.name, self.value)
        return '%s %s' % (self.__class__.__name__, self.name)

    def __repr__(self):
        return self.__str__()

    def __eq__(self, node):
        return bool(type(self) == type(
            node) and self.name == node.name and self.value == node.value and
                    self.child_nodes == node.child_nodes)

    def __hash__(self):
        return hash('%s%s%s%s' % (type(self), self.name, self.value,
                                  hash(tuple(self.child_nodes))))

    def getChildNodes(self):
        return [n for n in self.child_nodes if isinstance(n, ASTNode)]

    def append(self, node):
        if isinstance(node, list):
            self.extend(node)
        else:
            if type(node) is EatPrevious:
                del self.child_nodes[-1]
            else:
                try:
                    node.parent = self
                except AttributeError as e:
                    print(e, node)
                    raise
                self.child_nodes.append(node)

    def prepend(self, node):
        if isinstance(node, list):
            for n in reversed(node):
                self.child_nodes.insert(0, n)
        else:
            self.child_nodes.insert(0, node)

    # some classes override append() so just call down to that for now
    def extend(self, node_list):
        for n in node_list:
            self.append(n)

    def insert_before(self, marker_node, insert_node):
        """WARNING: If the same marker node appears more than once in the
        child_nodes list, the insert_node will be inserted before the first
        occurance."""
        try:
            idx = self.child_nodes.index(marker_node)
        except ValueError:
            raise ValueError("can't find child node %s in %s" %
                             (marker_node, self))
        # print ' '.join('insert_before', idx, id(self), self,
        #                id(marker_node), marker_node)
        insert_node.parent = self
        self.child_nodes.insert(idx, insert_node)

    def remove(self, node):
        self.replace(node, [])

    def replace(self, marker_node, insert_node_list):
        """Replace a node with another node or NodeList.

        If a NodeList is used, replace that child node with all of the
        nodes in the list. If it is just a basic node, replace the child
        node with that single node.
        """
        try:
            idx = self.child_nodes.index(marker_node)
        except ValueError:
            raise ValueError("can't find child node %s in %s" %
                             (marker_node, self))
        try:
            for n in reversed(insert_node_list):
                n.parent = self
                self.child_nodes.insert(idx, n)
        except TypeError:
            insert_node_list.parent = self
            self.child_nodes.insert(idx, insert_node_list)
        self.child_nodes.remove(marker_node)

    def copy(self, copy_children=True):
        if not copy_children:
            node = copy.copy(self)
            node.child_nodes = NodeList()
        else:
            node = copy.deepcopy(self)
        return node


class NodeList(list):

    def __init__(self, pos=None):
        # Position in the input string (measured in characters).
        self.pos = pos
        self.end_pos = None

    # note: need to iterate over a copy due to the way i modify the tree in-
    # place this is probably an indication that this approach is fundamentally
    # flawed
    def __iter__(self):
        return iter(list(list.__iter__(self)))

    def append(self, node):
        if isinstance(node, list):
            self.extend(node)
        else:
            list.append(self, node)

    def __hash__(self):
        # Equality is list equality.
        result = 0
        for item in self:
            result ^= hash(item)
        return result


class _ListNode(ASTNode):

    def __init__(self, parg_list=None, karg_list=None, pos=None):
        ASTNode.__init__(self, pos=pos)
        self.parg_list = parg_list
        self.karg_list = karg_list
        if parg_list:
            self.extend(parg_list)
        else:
            self.parg_list = []

        if karg_list:
            self.extend(karg_list)
        else:
            self.karg_list = []

    def __len__(self):
        return len(self.child_nodes)

    def __iter__(self):
        return iter(self.child_nodes)

    def __str__(self):
        return '%s %s' % (ASTNode.__str__(self),
                          ', '.join(str(n) for n in self.child_nodes))

    def __getitem__(self, item):
        return self.child_nodes[item]

    def get_arg_map(self):
        arg_map = {}
        for parameter_node in self.child_nodes:
            if not isinstance(parameter_node, ParameterNode):
                continue
            if parameter_node.default:
                arg_map[parameter_node.name] = parameter_node.default.value
            else:
                arg_map[parameter_node.name] = NoParameter
        return arg_map

    def get_arg_node_map(self):
        arg_map = {}
        for parameter_node in self.child_nodes:
            if not isinstance(parameter_node, ParameterNode):
                continue
            if parameter_node.default:
                arg_map[parameter_node.name] = parameter_node.default
            else:
                arg_map[parameter_node.name] = NoParameter
        return arg_map

    def get_parg_list(self):
        return self.parg_list


class ArgListNode(_ListNode):
    pass


class BinOpNode(ASTNode):

    def __init__(self, operator, left, right, pos=None):
        ASTNode.__init__(self, pos=pos)
        self.operator = operator
        self.left = left
        self.right = right

    def replace(self, node, replacement_node):
        if self.left is node:
            self.left = replacement_node
        elif self.right is node:
            self.right = replacement_node
        else:
            raise Exception("neither left nor right expression matches target")

    def __str__(self):
        return '%s (%s %s %s)' % (self.__class__.__name__, self.left,
                                  self.operator, self.right)

    def __eq__(self, node):
        return bool(type(self) == type(node) and self.operator == node.operator
                    and self.left == node.left and self.right == node.right)

    def __hash__(self):
        return hash('%s%s%s%s' % (type(self), self.operator, hash(self.left),
                                  hash(self.right)))


# fixme: are both BinOpNode and BinOpExpressionNode needed?
class BinOpExpressionNode(BinOpNode):
    pass


class AssignNode(BinOpNode):

    def __init__(self, left, right, pos=None):
        BinOpNode.__init__(self, '=', left, right, pos=pos)


class BreakNode(ASTNode):
    pass


class DoNode(ASTNode):
    """This node allows for an expression to be treated like a statement.

    The result of the expression is ignored and not written to the buffer."""

    def __init__(self, expression=None, pos=None):
        ASTNode.__init__(self, pos=pos)
        self.expression = expression

    def replace(self, node, replacement_node):
        if self.expression is node:
            self.expression = replacement_node
        else:
            raise Exception("expression doesn't match replacement")

    def __eq__(self, node):
        return bool(type(self) == type(node) and self.expression ==
                    node.expression and self.child_nodes == node.child_nodes)

    def __hash__(self):
        return hash('%s%s%s' % (type(self), hash(self.expression),
                                hash(tuple(self.child_nodes))))

    def __str__(self):
        return '%s expr:%s' % (self.__class__.__name__, self.expression)


class SanitizedState(object):
    """An enum of sanitization states of the return value for a function call.

    SANITIZED_STRING: We know for sure that the return value is already filtered
    and is a string. Threrefore, we wrap the call in a SanitizedPlaceholder
    directly.

    SANITIZED: We know for sure that the return value is already filtered.
    Therefore, we wrap the call in mark_as_sanitized.

    UNKNOWN: We aren't sure if the return type is filtered. Therefore, we
    check for a skip_filter annotation on the function at runtime.

    UNSANITIZED: We are sure that the function does not return a filtered value
    so we should not do anything with the return value.

    NOT_OUTPUTTED: We know that the function output will not be stored in a
    variable or passed into another function. Therefore, we do not need to do
    anything with the return value.

    OUTPUTTED_IMMEDIATELY: We know that the result of the function will be used
    immediately, so the skip_filter information on the function will be
    sufficient and there is no need to wrap the result in a
    SanitizedPlaceholder.
    """
    SANITIZED_STRING = 1
    SANITIZED = 2
    UNKNOWN = 3
    UNSANITIZED = 4
    NOT_OUTPUTTED = 5
    OUTPUTTED_IMMEDIATELY = 6


class CallFunctionNode(ASTNode):

    def __init__(self, expression=None, arg_list=None, pos=None):
        ASTNode.__init__(self, pos=pos)
        self.expression = expression
        # Whether we're calling into a library template.
        self.library_function = False
        # What the current sanitized state is of the function call.
        # See SanitizedState.
        self.sanitization_state = SanitizedState.UNKNOWN
        if arg_list:
            self.arg_list = arg_list
        else:
            self.arg_list = ArgListNode()

    def replace(self, node, replacement_node):
        if self.expression is node:
            self.expression = replacement_node
        else:
            raise Exception("expression doesn't match replacement")

    def __eq__(self, node):
        return bool(type(self) == type(node) and self.library_function ==
                    node.library_function and self.expression == node.expression
                    and self.arg_list == node.arg_list and (
                        self.sanitization_state == node.sanitization_state
                    ) and self.child_nodes == node.child_nodes)

    def __hash__(self):
        return hash('%s%s%s%s%s' % (type(self), hash(self.library_function),
                                    hash(self.expression), hash(self.arg_list),
                                    hash(tuple(self.child_nodes))))

    def __str__(self):
        return '%s expr:%s arg_list:%s' % (self.__class__.__name__,
                                           self.expression, self.arg_list)


# encapsulate the idea that you want to write this to an output stream
# this is sort of an implicit function call, so the hierarchy makes some sense
class BufferWrite(CallFunctionNode):

    def append_text_node(self, node):
        if not (isinstance(node, BufferWrite) and
                isinstance(node.expression, LiteralNode)):
            raise Exception('node type mismatch')
        self.expression.value += node.expression.value


class BufferExtend(CallFunctionNode):
    """Class to batch writes to the buffer."""
    pass


class CacheNode(CallFunctionNode):

    def __init__(self, expression=None, pos=None):
        ASTNode.__init__(self, '_cph%08X' % unsigned_hash(expression), pos=pos)
        self.expression = expression

    def __eq__(self, node):
        return bool(type(self) == type(node) and self.expression ==
                    node.expression)

    def __hash__(self):
        return hash('%s%s' % (type(self), hash(self.expression)))

    def __str__(self):
        return '%s expr:%s' % (self.__class__.__name__, self.expression)


class EchoNode(ASTNode):

    def __init__(self,
                 true_expression=None,
                 test_expression=None,
                 false_expression=None,
                 pos=None):
        ASTNode.__init__(self, pos=pos)
        self.true_expression = true_expression
        self.test_expression = test_expression
        self.false_expression = false_expression

    def replace(self, node, replacement_node):
        if self.true_expression is node:
            self.true_expression = replacement_node
        elif self.test_expression is node:
            self.test_expression = replacement_node
        elif self.false_expression is node:
            self.false_expression = replacement_node
        else:
            raise Exception("expression does not match target")


# dopey semi singleton - the nodes get cloned via deepcopy, so just
# make anything of this class identical
class __DefaultFilterFunction(object):

    def __eq__(self, o):
        return isinstance(o, self.__class__)

    def __hash__(self):
        return hash(self.__class__.__name__)


DefaultFilterFunction = __DefaultFilterFunction()


# encapsulate the idea that you want to run a filter over this expression
# this is sort of an implicit function call, so the hierarchy makes some sense
# again, in this case we want to preserve plenty of information and hierarchy
# for ease of optimization later on in the process
class FilterNode(ASTNode):

    def __init__(self,
                 expression=None,
                 filter_function_node=DefaultFilterFunction,
                 pos=None):
        ASTNode.__init__(self, pos=pos)
        self.expression = expression
        self.filter_function_node = filter_function_node

    def replace(self, node, replacement_node):
        if self.expression is node:
            self.expression = replacement_node
        else:
            raise Exception("expression doesn't match replacement")

    def __eq__(self, node):
        return bool(type(self) == type(
            node) and self.expression == node.expression and
                    self.filter_function_node == node.filter_function_node)

    def __hash__(self):
        return hash('%s%s%s' % (type(self), hash(self.expression),
                                hash(self.filter_function_node)))

    def __str__(self):
        return '%s expr:%s %s' % (self.__class__.__name__, self.expression,
                                  hash(self))


class CommentNode(ASTNode):
    pass


class ContinueNode(ASTNode):
    pass


class DefNode(ASTNode):

    def __init__(self, *pargs, **kargs):
        ASTNode.__init__(self, *pargs, **kargs)
        self.parameter_list = ParameterListNode()

    def __str__(self):
        return '%s name:%s parameter_list:%s' % (self.__class__.__name__,
                                                 self.name, self.parameter_list)

    def __eq__(self, other):
        return bool(type(self) == type(other) and self.name == other.name and
                    self.parameter_list == other.parameter_list and
                    self.child_nodes == other.child_nodes)

    def __hash__(self):
        return hash('%s%s%s%s' % (type(self), hash(self.name),
                                  hash(tuple(self.parameter_list)),
                                  hash(tuple(self.child_nodes))))


class BlockNode(DefNode):
    pass


class MacroNode(DefNode):
    pass


class DictLiteralNode(ASTNode):

    def append(self, node):
        self.child_nodes.append(node)

    def insert_before(self, marker_node, insert_node):
        raise ValueError('unsupported insert_before')

    def replace(self, marker_node, insert_node_list):
        insert_node = insert_node_list
        try:
            insert_node.parent = self
        except AttributeError:
            raise AttributeError('no parent attribute on %s' % insert_node)
        for idx, (key_node, value_node) in enumerate(self.child_nodes):
            if marker_node == key_node:
                self.child_nodes[idx] = (insert_node, value_node)
                break
            elif marker_node == value_node:
                self.child_nodes[idx] = (key_node, insert_node)
                break
        else:
            raise ValueError("can't find child node %s in %s" %
                             (marker_node, self))


class ExpressionListNode(_ListNode):
    pass


class ForNode(ASTNode):

    def __init__(self, target_list=None, expression_list=None, pos=None):
        ASTNode.__init__(self, pos=pos)
        if target_list:
            self.target_list = target_list
        else:
            self.target_list = TargetListNode()
        if expression_list:
            self.expression_list = expression_list
        else:
            self.expression_list = ExpressionListNode()
        self.scope = Scope('ForNode')
        self.loop_variant_set = None

    def __str__(self):
        return (
            '%s target_list:%s expr_list:%s' %
            (self.__class__.__name__, self.target_list, self.expression_list))


class StripLinesNode(ASTNode):
    """These are thrown away by the analyzer and do not make it to the
    codegen stage."""


class FunctionNode(ASTNode):

    def __init__(self, *pargs, **kargs):
        ASTNode.__init__(self, *pargs, **kargs)

        # PSEUDOCODE moved to codegen phase
        self.parameter_list = ParameterListNode()
        self.scope = Scope('Function')

    def __str__(self):
        return '%s %s parameter_list:%r' % (self.__class__.__name__, self.name,
                                            self.parameter_list)


class GetAttrNode(ASTNode):

    def __init__(self, expression, name, pos=None):
        ASTNode.__init__(self, pos=pos)
        self.expression = expression
        self.name = name

    def __eq__(self, node):
        return bool(type(self) == type(node) and self.name == node.name and
                    self.expression == node.expression)

    def __hash__(self):
        return hash('%s%s%s' %
                    (type(self), hash(self.name), hash(self.expression)))

    def __str__(self):
        return '%s expr:%s . name:%s' % (self.__class__.__name__,
                                         self.expression, self.name)

    def getChildNodes(self):
        child_nodes = self.expression.getChildNodes()
        child_nodes.append(self.expression)
        if isinstance(self.name, ASTNode):
            child_nodes.append(self.name)
        return child_nodes

    def replace(self, node, replacement_node):
        if self.expression is node:
            self.expression = replacement_node
        else:
            raise Exception("expression doesn't match replacement")


class GetUDNNode(GetAttrNode):
    pass


class IdentifierNode(ASTNode):
    # all subclasses of IdentifierNode should be treated as equivalent
    def __eq__(self, node):
        return bool(isinstance(node, IdentifierNode) and self.name == node.name)

    def __hash__(self):
        return hash(self.name)


class TemplateMethodIdentifierNode(IdentifierNode):
    pass


class IfNode(ASTNode):

    def __init__(self, test_expression=None, pos=None):
        ASTNode.__init__(self, pos=pos)
        self.test_expression = test_expression
        self.else_ = ElseNode(self)
        self.scope = Scope('If')

    def replace(self, node, replacement_node):
        if self.test_expression is node:
            self.test_expression = replacement_node
        else:
            ASTNode.replace(self, node, replacement_node)

    def __str__(self):
        return '%s test_expr:%s\nScope:%s\nelse:\n  %s' % (
            self.__class__.__name__, self.test_expression, self.scope,
            self.else_)

    def __eq__(self, other):
        return bool(type(self) == type(
            other) and self.test_expression == other.test_expression and
                    self.else_ == other.else_ and self.scope == other.scope and
                    self.child_nodes == other.child_nodes)

    def __hash__(self):
        return hash("%s%s%s%s%s" %
                    (type(self), hash(self.test_expression), hash(self.else_),
                     hash(self.scope), hash(self.child_nodes)))


class ElseNode(ASTNode):

    def __init__(self, parent=None, pos=None):
        ASTNode.__init__(self, pos=pos)
        self.parent = parent
        self.scope = Scope('Else')

    def __str__(self):
        return '%s %s' % (self.__class__.__name__, hash(self))


class ImplementsNode(ASTNode):
    pass


class AllowUndeclaredGlobalsNode(ASTNode):
    pass


class LooseResolutionNode(ASTNode):
    pass


class AllowRawNode(ASTNode):
    pass


class ImportNode(ASTNode):

    def __init__(self, module_name_list, library=False, pos=None):
        ASTNode.__init__(self, pos=pos)
        self.module_name_list = module_name_list
        self.library = library
        # in case you have a different target, save a copy of the
        # orginal name to use for dependency analysis
        self.source_module_name_list = module_name_list[:]

    def __eq__(self, node):
        return bool(type(self) == type(node) and self.library == node.library
                    and self.module_name_list == node.module_name_list)

    def __hash__(self):
        return hash('%s%s' % (type(self), hash(tuple(self.module_name_list))))

    def __str__(self):
        return ('%s module_name_list:%r' %
                (self.__class__.__name__, self.module_name_list))


# alpha break
class ExtendsNode(ImportNode):
    pass


class AbsoluteExtendsNode(ExtendsNode):
    pass


class FromNode(ImportNode):

    def __init__(self,
                 module_name_list,
                 identifier,
                 alias=None,
                 library=False,
                 pos=None):
        ImportNode.__init__(self, module_name_list, library=library, pos=pos)
        self.identifier = identifier
        self.alias = alias

    def __eq__(self, node):
        return bool(type(self) == type(
            node) and self.module_name_list == node.module_name_list and
                    self.identifier == node.identifier)

    def __hash__(self):
        return hash('%s%s%s' % (type(self), hash(tuple(self.module_name_list)),
                                self.identifier))

    def __str__(self):
        return ('%s module_name_list:%r identifier:%s' %
                (self.__class__.__name__, self.module_name_list,
                 self.identifier))


class ListLiteralNode(ASTNode):

    def __str__(self):
        return '%s nodes:%r' % (self.__class__.__name__, self.child_nodes)


class LiteralNode(ASTNode):

    def __init__(self, value, pos=None):
        ASTNode.__init__(self, pos=pos)
        self.value = value

    def __str__(self):
        return '%s value:%r' % (self.__class__.__name__, self.value)


class GlobalNode(ASTNode):
    pass


class ParameterNode(ASTNode):

    def __init__(self, name, default=None, pos=None):
        ASTNode.__init__(self, name, pos=pos)
        self.default = default

    def replace(self, node, replacement_node):
        if self.default is node:
            self.default = replacement_node
        else:
            raise Exception("default expression does not match target")

    def __str__(self):
        return '%s %s' % (ASTNode.__str__(self), self.default)

    def __eq__(self, node):
        return bool(type(self) == type(node) and self.name == node.name and
                    self.default == node.default)

    def __hash__(self):
        return hash('%s%s%s' % (type(self), self.name, hash(self.default)))


class AttributeNode(ParameterNode):
    pass


class FilterAttributeNode(AttributeNode):
    pass


class __NoParameter(object):

    def __repr__(self):
        return '<NoParameter>'


NoParameter = __NoParameter()


class ParameterListNode(_ListNode):
    pass


class PlaceholderNode(ASTNode):
    pass


class PlaceholderSubstitutionNode(ASTNode):

    def __init__(self, expression, parameter_list=None, pos=None):
        ASTNode.__init__(self, pos=pos)
        self.expression = expression
        if parameter_list is None:
            self.parameter_list = ParameterListNode()
        else:
            self.parameter_list = parameter_list

    def __str__(self):
        return '%s expr:%r %s' % (self.__class__.__name__, self.expression,
                                  self.parameter_list)


class ReturnNode(ASTNode):

    def __init__(self, expression, pos=None):
        ASTNode.__init__(self, pos=pos)
        self.expression = expression

    def __str__(self):
        return '%s expr:%r' % (self.__class__.__name__, self.expression)


class SliceNode(ASTNode):

    def __init__(self, expression, slice_expression, pos=None):
        ASTNode.__init__(self, pos=pos)
        self.expression = expression
        self.slice_expression = slice_expression

    def __str__(self):
        return (
            '%s expr:%s [ %s ]' %
            (self.__class__.__name__, self.expression, self.slice_expression))

    def __eq__(self, node):
        return bool(type(self) == type(
            node) and self.expression == node.expression and
                    self.slice_expression == node.slice_expression)

    def __hash__(self):
        return hash('%s%s%s' % (type(self), hash(self.expression),
                                hash(self.slice_expression)))

    def replace(self, node, replacement_node):
        if self.expression is node:
            self.expression = replacement_node
        elif self.slice_expression is node:
            self.slice_expression = replacement_node
        else:
            raise Exception("neither expression matches target")


class TargetNode(IdentifierNode):
    pass


class TargetListNode(_ListNode):
    pass


class TextNode(ASTNode):

    def __init__(self, value, pos=None):
        ASTNode.__init__(self, pos=pos)
        self.value = value

    def append_text_node(self, node):
        if not isinstance(node, TextNode):
            raise Exception('node type mismatch')
        self.value += node.value

    def __eq__(self, other):
        return bool(type(self) == type(other) and self.value == other.value)

    def __hash__(self):
        return hash("%s%s" % (type(self), self.value))


class WhitespaceNode(TextNode):

    def make_optional(self):
        return OptionalWhitespaceNode(self.value)


class NewlineNode(WhitespaceNode):
    pass


class OptionalWhitespaceNode(WhitespaceNode):
    pass


class FragmentNode(ASTNode):
    pass


class TemplateNode(ASTNode):
    __builtin_set = frozenset(dir(__builtins__))

    def __init__(self, classname=None, pos=None, **kargs):
        ASTNode.__init__(self, pos=pos, **kargs)
        self.source_path = None
        # fixme: need to get the classname from somewhere else
        self.classname = classname
        self.main_function = FunctionNode(name='main')
        self.main_function.parameter_list = ParameterListNode()
        self.main_function.parameter_list.append(ParameterNode(name='self'))
        self.encoding = 'utf-8'
        self.extends_nodes = NodeList()
        self.import_nodes = NodeList()
        self.from_nodes = NodeList()
        self.attr_nodes = NodeList()
        self.library = False
        self.implements = False
        self.allow_undeclared_globals = False
        self.use_loose_resolution = False
        self.baked = False
        self.explicitly_allow_raw = False
        self.global_placeholders = set()
        self.global_identifiers = set()
        self.cached_identifiers = set()
        self.template_methods = set()
        self.library_identifiers = set()
        self.trusted_module_identifiers = set()
        self.local_scope_identifiers = set()
        self.used_function_registry_identifiers = set()

    def __str__(self):
        return '%s\nimport:%s\nfrom:%s\nextends:%s\nmain:%s' % (
            self.__class__.__name__, self.import_nodes, self.from_nodes,
            self.extends_nodes, self.main_function)

    def has_identifier(self, name):
        return (name in self.__builtin_set or
                name in self.local_scope_identifiers or
                name in self.template_methods or
                name in self.trusted_module_identifiers or
                name in [node.name for node in self.attr_nodes])


class TupleLiteralNode(ASTNode):
    pass


class UnaryOpNode(ASTNode):

    def __init__(self, operator, expression, pos=None):
        ASTNode.__init__(self, pos=pos)
        self.operator = operator
        self.expression = expression

    def replace(self, node, replacement_node):
        if self.expression is node:
            self.expression = replacement_node
        else:
            raise Exception("expression does not match target")

    def __eq__(self, node):
        return bool(type(self) == type(node) and self.name == node.name and
                    self.expression == node.expression)

    def __hash__(self):
        return hash('%s%s%s' % (type(self), self.name, self.expression))


# save state related to scoping rules for code blocks
# this is kind of a hack. i am semi-emulating python scoping rules, which are
# a bit funky. probably, i should have defined a set of scope rules for
# templates and shoe-horned that into python
class Scope(object):

    def __init__(self, name=None):
        if name:
            self.name = name
        else:
            self.name = hex(id(self))
        self.local_identifiers = set()
        self.aliased_expression_map = OrderedDict()
        self.alias_name_set = ScopeSet()
        self.filtered_expression_map = {}
        self.hoisted_aliases = []
        # This set holds local identifiers that may or may not be defined
        # in the current scope. For example a variable defined in only the
        # if or the else clause of an if statement.
        self.partial_local_identifiers = set()
        # This set holds the local identifiers that are marked as dirty
        # (i.e. they have been used).
        self.dirty_local_identifiers = set()

    def __str__(self):
        return "<Scope %(name)s> %(alias_name_set)s" % vars(self)

    def __eq__(self, other):
        # TODO: Is this accurate enough?
        return (type(self) == type(other) and self.name == other.name and
                self.local_identifiers == other.local_identifiers)

    def __hash__(self):
        # local_identifiers is not hashable.
        return hash("%s%s" % (type(self), self.name))


class ScopeSet(set):
    pass
    #def add(self, o):
    #  set.add(self, o)


class OrderedDict(object):

    def __init__(self):
        self._dict = {}
        self._order = []

    def get(self, key, default=None):
        return self._dict.get(key, default)

    def __contains__(self, key):
        return key in self._dict

    def __getitem__(self, key):
        return self._dict[key]

    def __setitem__(self, key, value):
        if key not in self._dict:
            self._order.append(key)
            self._dict[key] = value

    def __delitem__(self, key):
        self._order.remove(key)
        del self._dict[key]

    def keys(self):
        return []

    def items(self):
        return []

    def iterkeys(self):
        return iter(self._order)

    def iteritems(self):
        for key in self._order:
            yield key, self._dict[key]

    def update(self, ordered_dict):
        for key, value in ordered_dict.items():
            self[key] = value

    def __str__(self):
        return str([(x, self._dict[x]) for x in self._order])


# this is sort of a hack to support optional white space nodes inside the
# parse tree.  the reality is that this probably requires a more complex
# parser, but we can get away with faking it for now.
def make_optional(node_list):
    try:
        if type(node_list[-1]) == WhitespaceNode:
            if (len(node_list) == 1 or type(node_list[-2]) == NewlineNode or
                    node_list[-2].statement or
                    not isinstance(node_list[-2],
                                   (TextNode, PlaceholderSubstitutionNode))):
                node_list[-1] = OptionalWhitespaceNode(node_list[-1].value)
    except IndexError:
        pass


# this is another hack to support line-wise stripping of white space nodes
# inside a #strip_lines directive.
def strip_whitespace(node_list, starts_new_line=True):
    # starts as optional only if we're at the beginning of a new line.
    optional = starts_new_line
    for i, node in enumerate(node_list):
        if isinstance(node, (OptionalWhitespaceNode, NewlineNode)):
            optional = True
            node_list[i] = OptionalWhitespaceNode('')
        elif isinstance(node, WhitespaceNode):
            if optional:
                node_list[i] = OptionalWhitespaceNode('')
            elif len(node_list) > i + 1 and isinstance(node_list[i + 1],
                                                       NewlineNode):
                optional = True
                node_list[i] = OptionalWhitespaceNode('')
        elif isinstance(node, TextNode):
            if optional:
                node.value = node.value.lstrip()
            optional = False
        elif not (node.statement or node.child_nodes or isinstance(
                node, CommentNode)):
            optional = False

        if optional and i > 0:
            prev_node = node_list[i - 1]
            if isinstance(prev_node, TextNode):
                prev_node.value = prev_node.value.rstrip()


def unsigned_hash(x):
    exp_hash = hash(x)
    if exp_hash < 0:
        exp_hash = -exp_hash | 0x80000000
    return exp_hash


def track_line_numbers(exempt_methods=()):
    """Class decorator: Change a parser class to track line numbers in
    the AST nodes returned by the given grammar rules.

    This will wrap all methods except the ones starting with an underscore
    or that are explicitly exempt.
    """

    def make_execute_rule(rule):

        def _execute_rule(self, *args, **kwargs):
            saved_position = self.file_position
            result = rule(self, *args, **kwargs)
            if isinstance(result, (ASTNode, NodeList)):
                result.pos = saved_position
                result.end_pos = self.file_position
            return result

        return _execute_rule

    def decorator(cls):
        for method_name in dir(cls):
            if method_name in exempt_methods or method_name.startswith('_'):
                continue
            rule = getattr(cls, method_name)
            if hasattr(rule, '__call__'):
                setattr(cls, method_name, make_execute_rule(rule))
        return cls

    return decorator

# This is a list of all of the nodes that are considered statements.
# They can not appear on the right hand of an assignment. This list is
# used to determine if a node should be moved out of the expression
# field of another node.
statement_nodes = (AbsoluteExtendsNode, AllowUndeclaredGlobalsNode, AssignNode,
                   BlockNode, BufferExtend, BufferWrite, CommentNode,
                   ContinueNode, DefNode, DoNode, EchoNode, ElseNode,
                   ExtendsNode, ForNode, FromNode, FunctionNode, GlobalNode,
                   IfNode, ImplementsNode, ImportNode, LooseResolutionNode,
                   MacroNode, ReturnNode, StripLinesNode, TextNode)



# Copyright 2007 The Spitfire Authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import sys
import traceback
import xml.dom.minidom

from spitfire.compiler import ast
from spitfire.compiler import util

enable_debug = False


def debug(func_name, dom_node):
    if not enable_debug:
        return
    if dom_node.attributes:
        print(func_name, dom_node.nodeName, list(dom_node.attributes.keys()))
    else:
        print(func_name, dom_node.nodeName)


class XHTML2AST(object):
    namespace = 'py'
    attr_op_namespace = 'pyattr'

    def build_template(self, filename):
        f = open(filename)
        data = f.read().decode('utf8')
        f.close()
        return self.parse(data)

    def parse(self, src_text):
        dom = xml.dom.minidom.parseString(src_text)
        template = ast.TemplateNode()
        template.extend(self.build_ast(dom))
        return template

    def build_ast(self, dom_node):
        debug('build_ast', dom_node)

        node_list = []

        if dom_node.attributes:
            # the key types have a precedence that needs to be preserved
            # www.zope.org/Documentation/Books/ZopeBook/2_6Edition/AppendixC.stx
            # since this is also how we scan the tree, on-error is included
            #
            # fixme: content/replace are mutually exclusive, that should
            # generate an error the thing is, the way we process things is a
            # little complicated, so the order is actually different - we might
            # handle something like omit-tag early on, but really only apply
            # it's implications later on
            op_precedence = [
                'omit-tag',
                'define',
                'condition',
                'repeat',
                'content',
                'content-html',
                'replace',
                'replace-html',
                'attributes',
                'on-error',
            ]

            # some of these operations can alter the output stream (most of them
            # really) - also, some don't exactly make sense to be on the same
            # object as a repeat - for instance, repeat->replace, whereas
            # repeat->attributes makes more sense

            # fixme: do I need keys() here? also, i think that attribute can be
            # None
            attr_name_list = list(dom_node.attributes.keys())
            processed_any_op = False
            for op in op_precedence:
                op_attr_name = '%s:%s' % (self.namespace, op)
                if dom_node.hasAttribute(op_attr_name):  # in attr_name_list:
                    op_handler = 'handle_%s' % op
                    op_handler = op_handler.replace('-', '_')
                    # print ' '.join("op_handler:", op_handler,
                    #                dom_node.nodeName,
                    #                dom_node.attributes.keys(),
                    #                processed_any_op)

                    node_list.extend(getattr(self, op_handler)(dom_node,
                                                               op_attr_name))
                    processed_any_op = True

            # process attribute namespace
            attr_output_ast = []
            attr_prune_list = []
            # this is horribly un-pythonic - i'm having Java flashbacks
            for i in range(dom_node.attributes.length):
                attr = dom_node.attributes.item(i)
                if attr.prefix == self.attr_op_namespace:
                    attr_prune_list.append(attr.localName)
                    attr_prune_list.append('%s:%s' % (self.attr_op_namespace,
                                                      attr.localName))
                    attr_output_ast.extend(self.make_attr_node(attr))
                    # print "attr_handler:", attr.prefix, attr.localName
                    #processed_any_op = True
            for attr_name in attr_prune_list:
                try:
                    dom_node.removeAttribute(attr_name)
                except xml.dom.NotFoundErr:
                    print("ignoring missing", attr_name)

            if not processed_any_op:
                node_list.extend(self.handle_default(dom_node,
                                                     attr_ast=attr_output_ast))
        else:
            node_list.extend(self.handle_default(dom_node))

        #for child in dom_node.childNodes:
        #  node_list.extend(self.build_ast(child))

        return node_list

    # attr_ast - allow injecting some ast nodes
    # fixme: feels like it could have a cleaner API
    def handle_default(self, dom_node, attr_ast=None):
        debug('handle_default', dom_node)
        node_list = []
        if dom_node.nodeType == xml.dom.minidom.Node.ELEMENT_NODE:
            node_list.extend(self.make_tag_node(dom_node, attr_ast=attr_ast))
            for child in dom_node.childNodes:
                node_list.extend(self.build_ast(child))
            node_list.extend(self.make_tag_node(dom_node, close=True))
        elif dom_node.nodeType == xml.dom.minidom.Node.TEXT_NODE:
            node_list.append(ast.TextNode(dom_node.nodeValue))
        elif dom_node.nodeType == xml.dom.minidom.Node.COMMENT_NODE:
            # node_list.append(ast.TextNode(dom_node.nodeValue))
            pass
        elif dom_node.nodeType == xml.dom.minidom.Node.DOCUMENT_NODE:
            for child in dom_node.childNodes:
                node_list.extend(self.build_ast(child))
        elif (dom_node.nodeType ==
              xml.dom.minidom.Node.PROCESSING_INSTRUCTION_NODE):
            if dom_node.nodeName == 'py-doctype':
                node_list.append(ast.TextNode(dom_node.nodeValue))
            else:
                raise Exception("unexepected processing instruction: %s" %
                                dom_node)
        else:
            raise Exception("unexepected node type: %s" % dom_node.nodeType)
        return node_list

    def make_tag_node(self, dom_node, close=False, attr_ast=None):
        debug("make_tag_node", dom_node)
        node_list = []
        node_name = dom_node.nodeName
        if close:
            if self.has_child_stuff(dom_node):
                node_list.append(ast.TextNode('</%(node_name)s>' % vars()))
        else:
            attr_text = ' '.join(['%s="%s"' % (key, value)
                                  for key, value in list(dom_node.attributes.items())
                                  if not key.startswith('py:')])
            # fixme: this is starting to look fugly - hard to maintain and error
            # prone
            if self.has_child_stuff(dom_node):
                if attr_text:
                    if attr_ast:
                        node_list.append(ast.TextNode(
                            '<%(node_name)s %(attr_text)s' % vars()))
                        node_list.extend(attr_ast)
                        node_list.append(ast.TextNode('>'))
                    else:
                        node_list.append(ast.TextNode(
                            '<%(node_name)s %(attr_text)s>' % vars()))
                else:
                    if attr_ast:
                        node_list.append(ast.TextNode('<%(node_name)s' % vars(
                        )))
                        node_list.extend(attr_ast)
                        node_list.append(ast.TextNode('>'))
                    else:
                        node_list.append(ast.TextNode('<%(node_name)s>' % vars(
                        )))
            else:
                if attr_text:
                    if attr_ast:
                        # print "XXX make_tag_node", dom_node.nodeName, attr_ast
                        node_list.append(ast.TextNode(
                            '<%(node_name)s %(attr_text)s' % vars()))
                        node_list.extend(attr_ast)
                        node_list.append(ast.TextNode(' />'))
                    else:
                        node_list.append(ast.TextNode(
                            '<%(node_name)s %(attr_text)s />' % vars()))
                else:
                    if attr_ast:
                        node_list.append(ast.TextNode('<%(node_name)s' % vars(
                        )))
                        node_list.extend(attr_ast)
                        node_list.append(ast.TextNode(' />'))
                    else:
                        node_list.append(ast.TextNode('<%(node_name)s />' %
                                                      vars()))

        omit_tag = getattr(dom_node, 'omit_tag', False)
        omit_tag_ast = getattr(dom_node, 'omit_tag_ast', None)
        if omit_tag:
            if omit_tag_ast:
                if_node = ast.IfNode(omit_tag_ast)
                if_node.extend(node_list)
                return [if_node]
            else:
                return []

        return node_list

    def make_attr_node(self, attr):
        node_list = []
        new_attr_name = attr.localName
        attr_ast = util.parse(attr.nodeValue, 'rhs_expression')
        node_list.append(ast.TextNode(' %(new_attr_name)s="' % vars()))
        # fixme: need to guarantee good output - escape sequences etc
        node_list.append(ast.PlaceholderSubstitutionNode(attr_ast))
        node_list.append(ast.TextNode('"'))
        return node_list

    def handle_define(self, dom_node, attr_name):
        node_list = []
        node_name = dom_node.nodeName
        # print "handle_define", node_name
        # fixme: this is a nasty temp hack, it will generate the correct code
        # for 1 define, but multiple expressions won't work
        expr_ast = util.parse(dom_node.getAttribute(attr_name), 'argument_list')
        dom_node.removeAttribute(attr_name)
        node_list.extend(expr_ast)
        node_list.extend(self.build_ast(dom_node))
        return node_list

    def handle_content(self, dom_node, attr_name):
        debug("handle_content", dom_node)
        #traceback.print_stack()
        expr_ast = util.parse(
            dom_node.getAttribute(attr_name), 'rhs_expression')
        dom_node.removeAttribute(attr_name)
        setattr(dom_node, 'has_child_stuff', True)
        node_list = []
        debug("handle_content start", dom_node)
        node_list.extend(self.make_tag_node(dom_node))
        node_list.append(ast.PlaceholderSubstitutionNode(expr_ast))
        debug("handle_content end", dom_node)
        node_list.extend(self.make_tag_node(dom_node, close=True))
        debug("handle_content return", dom_node)
        return node_list

    def handle_omit_tag(self, dom_node, attr_name):
        debug("handle_omit_tag", dom_node)
        node_list = []
        node_name = dom_node.nodeName
        raw_expression = dom_node.getAttribute(attr_name)
        if raw_expression:
            expr_ast = util.parse(raw_expression, 'argument_list')
        else:
            expr_ast = None

        dom_node.removeAttribute(attr_name)
        setattr(dom_node, 'omit_tag', True)
        setattr(dom_node, 'omit_tag_ast', expr_ast)
        return node_list

    def handle_replace(self, dom_node, attr_name):
        expr_ast = util.parse(
            dom_node.getAttribute(attr_name), 'rhs_expression')
        dom_node.removeAttribute(attr_name)
        return [ast.PlaceholderSubstitutionNode(expr_ast)]

    def has_child_stuff(self, dom_node):
        if getattr(dom_node, 'has_child_stuff', False):
            return True
        has_child_stuff = False
        for attr_name in ('py:content', 'py:replace',):
            if dom_node.hasAttribute(attr_name):
                has_child_stuff = True
                break
        else:
            has_child_stuff = bool(dom_node.childNodes)
        setattr(dom_node, 'has_child_stuff', has_child_stuff)
        return has_child_stuff

    def handle_repeat(self, dom_node, attr_name):
        debug("handle_repeat", dom_node)
        expr_pieces = dom_node.getAttribute(attr_name).split()
        dom_node.removeAttribute(attr_name)
        target = expr_pieces[0]
        expr_ast = util.parse(' '.join(expr_pieces[1:]), 'rhs_expression')
        node_list = []
        # hack - assumes python syntax
        fn = ast.ForNode(
            ast.TargetListNode([ast.IdentifierNode(
                "self.repeat['%s']" % target), ast.IdentifierNode(target)]),
            ast.ExpressionListNode([ast.CallFunctionNode(
                ast.IdentifierNode('enumerate'), ast.ArgListNode([expr_ast]))]))

        if self.has_child_stuff(dom_node):
            debug("has_child_stuff:", dom_node)
            fn.extend(self.build_ast(dom_node))
            #fn.append(self.make_tag_node(dom_node))
            #for n in dom_node.childNodes:
            #  fn.extend(self.build_ast(n))
        else:
            # print "no children"
            fn.extend(self.build_ast(dom_node))

        if (dom_node.previousSibling and dom_node.previousSibling.nodeType ==
                xml.dom.minidom.Node.TEXT_NODE and
                not dom_node.previousSibling.nodeValue.strip()):
            # inject the previous whitespace sibling to keep the output looking
            # ok fixme: a conditional is probably required here - you only want
            # to execute this if it's not the last execution of the loop
            fn.prepend(self.build_ast(dom_node.previousSibling))

            # now remove the previous sibling
            #print "node", dom_node
            #print "parent", dom_node.parentNode
            #print ' '.join("previous", dom_node.previousSibling,
            #               id(dom_node.previousSibling))
            #print "next", dom_node.nextSibling, id(dom_node.nextSibling)
            #dom_node.parentNode.removeChild(dom_node.previousSibling)
            node_list.append(ast.EatPrevious())

        node_list.append(fn)
        #fn.extend(self.make_tag_node(dom_node, close=True))
        return node_list

    def handle_condition(self, dom_node, attr_name):
        expr_ast = util.parse(
            dom_node.getAttribute(attr_name), 'rhs_expression')
        node_list = []
        if_node = ast.IfNode(expr_ast)
        node_list.append(if_node)
        if_node.append(self.make_tag_node(dom_node))
        for n in dom_node.childNodes:
            if_node.extend(self.build_ast(n))
        if_node.extend(self.make_tag_node(dom_node, close=True))
        return node_list

    def build_udn_path_ast(self, path):
        pieces = path.split('.')
        node = ast.PlaceholderNode(pieces[0])
        for piece in pieces[1:]:
            node = ast.GetUDNNode(node, piece)
        return node


if __name__ == '__main__':
    x2a = XHTML2AST()
    filename = sys.argv[1]
    tnode = x2a.build_template(filename)
    print(tnode)
    classname = spitfire.compiler.util.filename2classname(filename)
    src = spitfire.compiler.util.compile_ast(tnode, classname)
    print(src)
    module = spitfire.compiler.util.load_module_from_src(src, '<none>',
                                                         classname)
    tclass = getattr(module, classname)
    d = {
        'test_x': 'x var',
        'test_y': 'y var',
        'test_z': 'z var',
        'test_number_list': [1, 2, 3, 4, 5],
        'test_object_list': [{'id': 1,
                              'name': 'o1'},
                             {'id': 2,
                              'name': 'o2'},
                             {'id': 3,
                              'name': 'o3'},],
        'test_dict': {'key1': 1},
        'test_whitespaced_dict': {'key 1': 1},
        'test_range': range,
        'content_type': 'test/spitfire',
    }

    print(tclass(search_list=[d]).main())



# Copyright 2008 The Spitfire Authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import sys

import io as StringIO

from spitfire.compiler import analyzer
from spitfire.compiler import ast
from spitfire.compiler import util
from spitfire.compiler import visitor
from spitfire import text
from spitfire.compiler import util


# generate a reasonable substitute name from a raw placeholder node
def make_placeholder_name(placeholder_node):
    node_type = type(placeholder_node.expression)
    placeholder_name = ''
    if node_type is ast.PlaceholderNode:
        placeholder_name = placeholder_node.expression.name
    elif node_type is ast.CallFunctionNode:
        placeholder = placeholder_node.expression.expression
        if type(placeholder) is ast.PlaceholderNode:
            placeholder_name = placeholder.name

    placeholder_name = placeholder_name.upper()
    return placeholder_name


# create a translated version of the raw_msg while allowing placeholder
# expressions to pass through correctly
def make_i18n_message(raw_msg, macro_ast):
    # top should be a fragment and due to the limited syntax, we can more or
    # less scan this one level of nodes -- there are no nested i18n sections yet
    output = StringIO.StringIO()
    for i, n in enumerate(macro_ast.child_nodes):
        #print i, type(n), "start", n.start, "end", n.end
        #print "raw:", "'%s'" % raw_msg[n.start:n.end]

        if isinstance(n, ast.PlaceholderSubstitutionNode):
            raw_placeholder_expression = raw_msg[n.start:n.end]
            #output.write(make_placeholder_name(n))
            output.write(raw_placeholder_expression)
        else:
            output.write(text.i18n_mangled_message(n.value))
    return output.getvalue()


def macro_i18n(macro_node, arg_map, compiler):
    # fixme: parse the parameter list into something usable
    # macro_node.parameter_list

    # generate a fake translation for now to verify this is working
    # most apps will have to stub this part out somehow i think
    macro_content_ast = util.parse(macro_node.value, 'i18n_goal')
    i18n_msg = make_i18n_message(macro_node.value, macro_content_ast)
    i18n_msg_utf8 = i18n_msg.encode(sys.getdefaultencoding())
    #print "macro_content_ast"
    #print "orginal:", macro_node.value
    #print "i18n:", i18n_msg_utf8
    #visitor.print_tree(macro_content_ast)
    return i18n_msg


def macro_function_i18n(call_node, arg_map, compiler):
    # generate a fake translation for now to verify this is working
    # most apps will have to stub this part out somehow i think
    # in the context of a function, the goal is to replace a function with a
    # translated literal string. we have to do some shenanigans since Spitfire
    # doesn't parse repr(unicode)
    msg_arg_node = call_node.arg_list.parg_list[0]
    if not isinstance(msg_arg_node, ast.LiteralNode):
        raise analyzer.SemanticAnalyzerError(
            '$i18n argument "%s" must be a string literal' % msg_arg_node)
    i18n_msg = text.i18n_mangled_message(msg_arg_node.value)
    i18n_msg_utf8 = i18n_msg.encode(sys.getdefaultencoding())
    return "'%s'" % i18n_msg.replace("'", "\\'")



# Copyright 2007 The Spitfire Authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import copy
import logging
import os.path
import re

from spitfire.compiler import ast
from spitfire.compiler import analyzer
from spitfire.compiler import walker
from functools import reduce

builtin_names = dir(__builtins__)

_BINOP_INVALID_COUNT = 1000  # Any value > 0 will work.
_BINOP_INITIAL_COUNT = 0
_BINOP_FIRST_PASS = 1


# Utility functions for searching up the tree.
def _get_parent_node_by_pred(node, pred, search_current=False):
    """Find the first parent node that satisfies a predicate function."""
    if not search_current:
        node = node.parent
    while node is not None:
        if pred(node):
            return node
        node = node.parent
    return None


def _get_parent_node_by_type(node, node_type):
    """Find the first parent node that satisfies a specified type or types."""
    return _get_parent_node_by_pred(node, lambda n: isinstance(n, node_type))


def _get_parent_loop(node):
    return _get_parent_node_by_type(node, ast.ForNode)


def _get_parent_block(node):
    return _get_parent_node_by_type(
        node, (ast.FunctionNode, ast.ForNode, ast.IfNode, ast.ElseNode))


def _get_local_identifiers(node):
    local_identifiers = []
    partial_local_identifiers = []
    dirty_local_identifiers = []

    # search the parent scopes
    # fixme: should this be recursive?
    node = node.parent
    while node is not None:
        if isinstance(node, ast.ForNode):
            local_identifiers.extend(node.loop_variant_set)
            local_identifiers.extend(node.scope.local_identifiers)
            partial_local_identifiers.extend(
                node.scope.partial_local_identifiers)
            dirty_local_identifiers.extend(node.scope.dirty_local_identifiers)
        elif isinstance(node, ast.IfNode):
            local_identifiers.extend(node.scope.local_identifiers)
            partial_local_identifiers.extend(
                node.scope.partial_local_identifiers)
            dirty_local_identifiers.extend(node.scope.dirty_local_identifiers)
        elif isinstance(node, ast.ElseNode):
            # in this case, we don't want to go to the parent node, which is the
            # ast.IfNode - we want to go to the parent scope
            local_identifiers.extend(node.scope.local_identifiers)
            partial_local_identifiers.extend(
                node.scope.partial_local_identifiers)
            dirty_local_identifiers.extend(node.scope.dirty_local_identifiers)
            node = node.parent.parent
            continue
        elif isinstance(node, ast.FunctionNode):
            local_identifiers.extend(node.scope.local_identifiers)
            partial_local_identifiers.extend(
                node.scope.partial_local_identifiers)
            dirty_local_identifiers.extend(node.scope.dirty_local_identifiers)
            break
        node = node.parent
    return (frozenset(local_identifiers), frozenset(partial_local_identifiers),
            frozenset(dirty_local_identifiers))


def _get_identifiers_from_expression(node):
    """Find the IdentifierNodes present in an expression.

    This function searches through the nodes of an expression and returns a set
    of the IdentiferNodes that are present. This function doesn't traverse
    ast.GetAttrNode or any LiteralNodes such as ListLiteral or DictLiteral
    nodes.
    """
    return set([n
                for n in walker.flatten_tree(node)
                if isinstance(n, ast.IdentifierNode)])


def _is_clean(node, scope):
    """Determine if the node references any dirty identifiers in the scope.

    If there are any Identifiers in the node that are considered "dirty", this
    function will return False. This is because it is not safe to hoist a node
    that depends on modifications made within its parent scope.
    """
    dirty_identifiers = scope.dirty_local_identifiers
    node_identifiers = _get_identifiers_from_expression(node)
    # There are no dirty idenitifiers in the node being hoisted.
    return not node_identifiers & dirty_identifiers


def _get_common_aliased_expression_map(*scopes):
    """Get the common clean aliased expression map for multiple scopes."""
    if not scopes:
        return {}
    clean_key_sets = []
    for scope in scopes:
        clean_scope_keys = set()
        for alias in scope.aliased_expression_map.keys():
            if _is_clean(alias, scope):
                clean_scope_keys.add(alias)
        clean_key_sets.append(clean_scope_keys)

    common_clean_keys = reduce(lambda x, y: x & y, clean_key_sets)
    return dict([(key, scopes[0].aliased_expression_map[key])
                 for key in common_clean_keys])


# Utility functions for generating names.
def _generate_filtered_placeholder(node):
    """Given a node, generate a name for the cached filtered placeholder"""
    return '_fph%08X' % ast.unsigned_hash(node)


def _generate_cached_resolved_placeholder(node):
    """Given a node, generate a name for the cached udn placeholder"""
    return '_rudn%08X' % ast.unsigned_hash(node)


class _BaseAnalyzer(object):

    def __init__(self, ast_root, options, compiler):
        self.ast_root = ast_root
        self.options = options
        self.compiler = compiler
        self.unoptimized_node_types = set()
        # Used as a flag to determine how many binops we have analyzed.
        self.binop_count = _BINOP_INVALID_COUNT

    def optimize_ast(self):
        self.visit_ast(self.ast_root)
        if self.options.debug:
            print("unoptimized_node_types", self.unoptimized_node_types)
        return self.ast_root

    # build an AST node list from a single parse node
    # need the parent in case we are going to delete a node
    def visit_ast(self, node, parent=None):
        node.parent = parent
        method_name = 'analyze%s' % node.__class__.__name__
        method = getattr(self, method_name, self.default_optimize_node)
        if method_name in self.compiler.debug_flags:
            print(method_name, node)
        return method(node)

    def skip_analyze_node(self, node):
        return

    analyzeLiteralNode = skip_analyze_node
    analyzeIdentifierNode = skip_analyze_node
    analyzeTargetNode = skip_analyze_node

    def default_optimize_node(self, node):
        # print "default_optimize_node", type(node)
        self.unoptimized_node_types.add(type(node))
        return

    # this function has some rules that are a bit unclean - you aren't actually
    # looking for the 'parent' scope, but one you might insert nodes into. for
    # instance, you skip over a ast.ForNode so that optimizetions are inserted
    # in a loop-invariant fashion.
    def get_parent_scope(self, node):
        node_stack = [node]
        node = node.parent
        while node is not None:
            if type(node) == ast.FunctionNode:
                return node.scope
            elif type(node) == ast.IfNode:
                # elements of the test clause need to reference the next scope
                # "up" - usually the function, but could be another conditional
                # block fixme: if we ever implement "elif" this will have to get
                # fixed up
                if node_stack[-1] != node.test_expression:
                    return node.scope
            elif type(node) == ast.ElseNode:
                return node.scope
            elif type(node) == ast.ForNode:
                if node_stack[-1] != node.expression_list:
                    return node.scope
            node_stack.append(node)
            node = node.parent
        self.compiler.error(
            analyzer.SemanticAnalyzerError("expected a parent function"),
            pos=node_stack[-1].pos)

    def get_insert_block_and_point(self, node):
        original_node = node
        insert_marker = node
        node = node.parent
        while node is not None:
            if isinstance(node, (ast.FunctionNode, ast.ForNode, ast.IfNode,
                                 ast.ElseNode)):
                if insert_marker in node.child_nodes:
                    return node, insert_marker

            insert_marker = node
            node = node.parent
        self.compiler.error(
            analyzer.SemanticAnalyzerError("expected a parent block"),
            pos=node.pos)

    def replace_in_parent_block(self, node, new_node):
        insert_block, insert_marker = self.get_insert_block_and_point(node)
        insert_block.replace(insert_marker, new_node)

    def reanalyzeConditionalNode(self, conditional_node):
        if (not self.options.hoist_conditional_aliases and
                not self.options.cache_filtered_placeholders):
            return

        parent_node = conditional_node
        parent_block, insertion_point = self.get_insert_block_and_point(
            conditional_node)

        if self.options.hoist_conditional_aliases:
            #print "reanalyzeConditionalNode", conditional_node
            #print "  parent_block", parent_block
            #print "  parent_scope", parent_block.scope
            # NOTE: need to iterate over items, in case we modify something
            items = list(conditional_node.scope.aliased_expression_map.items())
            for alias_node, alias in items:
                #print "  check alias:", alias
                #print "    alias_node:", alias_node
                assign_alias_node = ast.AssignNode(alias,
                                                   alias_node,
                                                   pos=alias_node.pos)
                if alias_node in parent_block.scope.aliased_expression_map:
                    if self._is_condition_invariant(alias_node,
                                                    conditional_node):
                        #print "  hoist:", assign_alias_node
                        self.hoist(conditional_node, parent_block,
                                   insertion_point, alias_node,
                                   assign_alias_node)

    def reanalyzeLoopNode(self, loop_node):
        if not self.options.hoist_loop_invariant_aliases:
            return

        parent_block, insertion_point = self.get_insert_block_and_point(
            loop_node)
        # NOTE: need to iterate over items, in case we modify something
        for alias_node, alias in list(loop_node.scope.aliased_expression_map.items()):
            assign_alias = ast.AssignNode(alias, alias_node, pos=alias_node.pos)
            if alias_node in parent_block.scope.aliased_expression_map:
                if self._is_loop_invariant(alias_node, loop_node):
                    self.hoist(loop_node, parent_block, insertion_point,
                               alias_node, assign_alias)
            else:
                # if this alias is not already used in the parent scope, that's
                # ok, hoist it if it's loop invariant
                if self._is_loop_invariant(alias_node, loop_node):
                    loop_node.remove(assign_alias)
                    parent_block.insert_before(loop_node, assign_alias)
                    parent_block.scope.hoisted_aliases.append(alias_node)

    def _is_condition_invariant(self, node, conditional_node):
        """The _is_condition_invariant_legacy function is broken, but seems to
        be correct in some cases. Out of fear and redundancy, in order for
        something to be hoisted, it must pass the old and new tests.
        """
        return (self._is_condition_invariant_legacy(node, conditional_node) and
                _is_clean(node, conditional_node.scope))

    def _is_condition_invariant_legacy(self, node, conditional_node):
        node_dependency_set = self.get_node_dependencies(node)
        condition_invariant = (
            not node_dependency_set & conditional_node.scope.local_identifiers)
        #print "is_condition_invariant:", condition_invariant
        #print "  locals:", conditional_node.scope.local_identifiers
        #print "  deps:", node_dependency_set
        return condition_invariant

    def _is_loop_invariant(self, node, loop_node):
        """The _is_loop_invariant_legacy function is broken, but seems to be
        correct in some cases. Out of fear and redundancy, in order for
        something to be hoisted, it must pass the old and new tests.
        """
        return (self._is_loop_invariant_legacy(node, loop_node) and
                _is_clean(node, loop_node.scope))

    def _is_loop_invariant_legacy(self, node, loop_node):
        node_dependency_set = self.get_node_dependencies(node)
        #     print "is loop invariant node:", node
        #     for x in node_dependency_set:
        #       print "  dep:", x

        # find dependencies within the loop node but outside the node we're
        # checking
        node_dependency_set_except_node_tree = node_dependency_set - set(
            walker.flatten_tree(node))
        dependencies_within_loop = set(walker.flatten_tree(
            loop_node)).intersection(node_dependency_set_except_node_tree)
        depends_on_loop_variants = bool(loop_node.loop_variant_set.intersection(
            node_dependency_set))

        # TODO: Disabling warnings for now. They are useless witout
        # filenames. Also need to make sure all these cases are valid.

        # if not depends_on_loop_variants and dependencies_within_loop:
        #   # we can't assume this is invariant because it depends on other
        #   # nodes inside the loop. eventually we should hoist out both the
        #   # node and its dependencies.
        #   dependency_nodes = '\n'.join('  %s' % node.parent
        #                                for node in dependencies_within_loop)
        #   logging.warning("Cannot hoist possible loop invariant: %s.", node)
        #   logging.warning("Please move following dependencies "
        #                   "out of the loop:\n%s", dependency_nodes)

        return not depends_on_loop_variants and not dependencies_within_loop

    def get_node_dependencies(self, node):
        node_dependency_set = set(walker.flatten_tree(node))
        parent_block = _get_parent_block(node)

        for n in list(node_dependency_set):
            # when this is an identifier, you need to check all of the potential
            # the dependencies for that symbol, which means doing some crawling
            if isinstance(n, ast.IdentifierNode):
                identifier = n
                parent_block_to_check = parent_block
                while parent_block_to_check:
                    for block_node in parent_block_to_check.child_nodes:
                        if isinstance(block_node, ast.AssignNode):
                            if block_node.left == identifier:
                                node_dependency_set.update(
                                    self.get_node_dependencies(
                                        block_node.right))
                                parent_block_to_check = None
                                break
                        elif isinstance(block_node, ast.IfNode):
                            # if you encounter a conditional in your chain, you
                            # depend on any dependencies of the condition itself
                            # FIXME: calling
                            # get_node_dependencies(block_node.test_expression)
                            # causes an infinite loop, but that is probably the
                            # correct way forward to address the dependency
                            # chain
                            node_dependency_set.update(walker.flatten_tree(
                                block_node.test_expression))
                    else:
                        parent_block_to_check = _get_parent_block(
                            parent_block_to_check)
            #elif isinstance(n, (ast.GetUDNNode, ast.FilterNode)):
            #  node_dependency_set.update(
            #    self.get_node_dependencies(node.expression))
            #print "get_node_dependencies", node
            #print "  deps:", node_dependency_set
        return node_dependency_set


class OptimizationAnalyzer(_BaseAnalyzer):

    def analyzeParameterNode(self, parameter):
        self.visit_ast(parameter.default, parameter)
        return

    def analyzeTemplateNode(self, template):
        # If we have a function registry, add in the nodes that we determined
        # are used from the analyzer stage before we begin optimizing.
        for alias in sorted(self.compiler.function_name_registry):
            if alias not in template.used_function_registry_identifiers:
                continue
            fq_name, method = self.compiler.function_name_registry[alias]
            fq_name_parts = fq_name.split('.')
            self.ast_root.from_nodes.append(ast.FromNode(
                [ast.IdentifierNode(x) for x in fq_name_parts[:-1]
                ], ast.IdentifierNode(fq_name_parts[-1]), ast.IdentifierNode(
                    alias)))

        for n in template.from_nodes:
            if n.alias:
                template.global_identifiers.add(n.alias)
            else:
                template.global_identifiers.add(n.identifier)

        # scan extends for dependencies
        # this allows faster calling of template functions - we could also
        # tune ast.BufferWrite calls for these nodes
        if self.options.use_dependency_analysis:
            for n in template.extends_nodes:
                path = os.path.join(
                    *[ident_node.name
                      for ident_node in n.source_module_name_list])
                template_function_names = get_template_functions(
                    self.compiler.include_path, path)
                template.template_methods.update(template_function_names)

        self.visit_ast(template.main_function, template)
        for n in template.child_nodes:
            self.visit_ast(n, template)

    def analyzeFunctionNode(self, function):
        function.scope.local_identifiers.update([ast.IdentifierNode(
            n.name) for n in function.parameter_list])
        # Some binops optimzations can only be done in functions.
        self.binop_count = _BINOP_INITIAL_COUNT
        for n in function.child_nodes:
            self.visit_ast(n, function)
        self.binop_count = _BINOP_INVALID_COUNT

    def analyzeForNode(self, for_node):
        self.visit_ast(for_node.target_list, for_node)
        for_node.loop_variant_set = set(for_node.target_list.flat_list)
        self.visit_ast(for_node.expression_list, for_node)
        for n in for_node.child_nodes:
            self.visit_ast(n, for_node)

    def analyzeAssignNode(self, node):
        scope = self.get_parent_scope(node)
        local_identifiers, _, dirty_identifiers = _get_local_identifiers(node)
        # This is an assignment at an index.
        if isinstance(node.left, ast.SliceNode):
            _identifier = ast.IdentifierNode(node.left.expression.name,
                                             pos=node.pos)
            scope.dirty_local_identifiers.add(_identifier)
            if _identifier not in local_identifiers:
                self.compiler.error(
                    analyzer.SemanticAnalyzerError(
                        'Expression %s being indexed must be defined before use'
                        % _identifier.name),
                    pos=node.pos)
        else:
            _identifier = ast.IdentifierNode(node.left.name, pos=node.pos)
            alias_name = _generate_filtered_placeholder(_identifier)
            if alias_name in scope.alias_name_set:
                if self.options.double_assign_error:
                    self.compiler.error(
                        analyzer.SemanticAnalyzerError(
                            'Multiple assignment of %s' % _identifier.name),
                        pos=node.pos)
                else:
                    self.compiler.warn('Multiple assignment of %s' %
                                       _identifier.name,
                                       pos=node.pos)
            scope.local_identifiers.add(_identifier)
        # note: this hack is here so you can partially analyze alias nodes
        # without double-processing
        if node.right:
            self.visit_ast(node.right, node)

    def analyzeExpressionListNode(self, expression_list_node):
        for n in expression_list_node:
            self.visit_ast(n, expression_list_node)

    def analyzeTargetListNode(self, target_list_node):
        flat_list = []
        for n in target_list_node:
            self.visit_ast(n, target_list_node)
            if type(n) == ast.TargetListNode:
                flat_list.extend(n.flat_list)
            else:
                flat_list.append(n)
        target_list_node.flat_list = flat_list

#   def analyzeParameterListNode(self, parameter_list_node):
#     flat_list = []
#     for n in parameter_list_node:
#       flat_list.append(n)
#     target_list_node.flat_list = flat_list

    def analyzeArgListNode(self, arg_list_node):
        scope = self.get_parent_scope(arg_list_node)
        for n in arg_list_node:
            # If an identifier is passed into a function and we are not in a
            # filter node, mark it as dirty. Filter nodes are always written
            # out, therefore we don't consider the final call a modification.
            # This assumption is predictaed on the fact that you can't modify a
            # variable once it has been written.
            if not _get_parent_node_by_type(arg_list_node, ast.FilterNode):
                if isinstance(n, ast.PlaceholderNode):
                    scope.dirty_local_identifiers.add(ast.IdentifierNode(
                        n.name))
                if isinstance(n, ast.ParameterNode) and isinstance(
                        n.default, ast.PlaceholderNode):
                    scope.dirty_local_identifiers.add(ast.IdentifierNode(
                        n.default.name))
            self.visit_ast(n, arg_list_node)

    def analyzeTupleLiteralNode(self, tuple_literal_node):
        for n in tuple_literal_node.child_nodes:
            self.visit_ast(n, tuple_literal_node)

    def analyzeDictLiteralNode(self, dict_literal_node):
        for key_node, value_node in dict_literal_node.child_nodes:
            self.visit_ast(key_node, dict_literal_node)
            self.visit_ast(value_node, dict_literal_node)

    def analyzeListLiteralNode(self, list_literal_node):
        for n in list_literal_node.child_nodes:
            self.visit_ast(n, list_literal_node)

    def analyzeDoNode(self, do_node):
        self.visit_ast(do_node.expression, do_node)

    def analyzeCallFunctionNode(self, function_call):
        # If the ast.CallFunctionNode is in the test expression of an
        # ast.IfNode, do not wrap the function with a SanitizedPlaceholder.
        def is_in_test_expression(node):
            return (isinstance(node.parent, ast.IfNode) and
                    node == node.parent.test_expression)

        if _get_parent_node_by_pred(function_call,
                                    is_in_test_expression,
                                    search_current=True):
            function_call.sanitization_state = ast.SanitizedState.NOT_OUTPUTTED
        # If the ast.CallFunctionNode is in a ast.DoNode, do not wrap the
        # function with a SanitizedPlaceholder.
        if _get_parent_node_by_type(function_call, ast.DoNode):
            function_call.sanitization_state = ast.SanitizedState.NOT_OUTPUTTED

        self.visit_ast(function_call.expression, function_call)
        self.visit_ast(function_call.arg_list, function_call)

    # NOTE: these optimizations are disabled because the optimizer has a
    # tendency to "over-hoist" code inside a ast.CacheNode and you end up doing
    # *more* work
    def analyzeCacheNode(self, cache_node):
        cache_placeholders = self.options.cache_resolved_placeholders
        cache_udn_expressions = self.options.cache_resolved_udn_expressions
        cache_filtered_placeholders = self.options.cache_filtered_placeholders
        self.options.cache_resolved_placeholders = False
        self.options.cache_resolved_udn_expressions = False
        self.options.cache_filtered_placeholders = False

        self.visit_ast(cache_node.expression, cache_node)

        self.options.cache_resolved_placeholders = cache_placeholders
        self.options.cache_resolved_udn_expressions = cache_udn_expressions
        self.options.cache_filtered_placeholders = cache_filtered_placeholders

    def analyzeBufferWrite(self, buffer_write):
        self.visit_ast(buffer_write.expression, buffer_write)
        # template functions output text - don't format them as strings
        if (isinstance(buffer_write.expression, ast.BinOpNode) and
                buffer_write.expression.operator == '%' and
                isinstance(buffer_write.expression.right, ast.CallFunctionNode)
                and isinstance(buffer_write.expression.right.expression,
                               ast.TemplateMethodIdentifierNode)):
            buffer_write.replace(buffer_write.expression,
                                 buffer_write.expression.right)

    def analyzeEchoNode(self, node):
        for n in (node.test_expression, node.true_expression,
                  node.false_expression):
            if n:
                self.visit_ast(n, node)

    def analyzeFilterNode(self, filter_node):
        self.visit_ast(filter_node.expression, filter_node)

        if isinstance(filter_node.expression, ast.CallFunctionNode):
            # If a ast.FilterNode has a ast.CallFunctionNode as an expression,
            # let the filter node handle determining if filtering should occur,
            # rather than the sanitization wrapper.
            filter_node.expression.sanitization_state = (
                ast.SanitizedState.OUTPUTTED_IMMEDIATELY)

        if (isinstance(filter_node.expression, ast.CallFunctionNode) and
                isinstance(filter_node.expression.expression,
                           ast.TemplateMethodIdentifierNode)):
            filter_node.parent.replace(filter_node, filter_node.expression)
            return

        # A ast.CallFunctionNode will require passing in both the value and the
        # function to the filter_function. If the ast.CallFunctionNode's
        # expression is a ast.GetUDNNode, we can avoid looking up the attribute
        # twice by caching its value. This is also true if the
        # ast.CallFunctionNode's expression is an ast.IdentifierNode with a '.'
        # in the name.
        if isinstance(filter_node.expression, ast.CallFunctionNode):
            fn_node = filter_node.expression
            if (self.options.cache_resolved_udn_expressions and
                    isinstance(fn_node.expression, ast.IdentifierNode) and
                    '.' in fn_node.expression.name):
                scope = self.get_parent_scope(filter_node)
                udn_node = fn_node.expression
                # For some udn style node, create a cached variable. ex.
                # _rudn12345
                alias_name = _generate_cached_resolved_placeholder(udn_node)
                alias = ast.IdentifierNode(alias_name, pos=filter_node.pos)
                scope.local_identifiers.add(alias)
                # Create an assignment node that assigns the udn node to the
                # ast.IdentifierNode.
                assign_alias = ast.AssignNode(alias,
                                              udn_node,
                                              pos=filter_node.pos)
                insert_block, insert_marker = self.get_insert_block_and_point(
                    filter_node)
                # Insert the assignment before the ast.FilterNode
                insert_block.insert_before(insert_marker, assign_alias)
                # Replace the udn node in the ast.CallFunctionNode with the
                # alias.
                filter_node.expression.replace(udn_node, alias)

        if self.options.cache_filtered_placeholders:
            # NOTE: you *must* analyze the node before putting it in a dict
            # otherwise the definition of hash and equivalence will change and
            # the node will not be found due to the sketchy custom hash function
            scope = self.get_parent_scope(filter_node)
            alias = scope.aliased_expression_map.get(filter_node)

            if not alias:
                alias_name = _generate_filtered_placeholder(
                    filter_node.expression)
                if alias_name in scope.alias_name_set:
                    print("duplicate alias_name", alias_name)
                    print("scope", scope)
                    print("scope.alias_name_set", scope.alias_name_set)
                    print(" ".join("scope.aliased_expression_map",
                                   scope.aliased_expression_map))
                    return

                alias = ast.IdentifierNode(alias_name, pos=filter_node.pos)
                scope.alias_name_set.add(alias_name)
                scope.aliased_expression_map[filter_node] = alias
                assign_alias = ast.AssignNode(alias,
                                              filter_node,
                                              pos=filter_node.pos)

                insert_block, insert_marker = self.get_insert_block_and_point(
                    filter_node)
                insert_block.insert_before(insert_marker, assign_alias)

            filter_node.parent.replace(filter_node, alias)

    def _placeholdernode_replacement(self, placeholder, local_var,
                                     cached_placeholder, local_identifiers):
        """This function tries to replace a ast.PlaceholderNode with a node type
        that does not need to be resolved such as an ast.IdentifierNode or a
        cached placeholder.
        """
        if local_var in local_identifiers:
            placeholder.parent.replace(placeholder, local_var)
        elif placeholder.name in self.ast_root.template_methods:
            placeholder.parent.replace(
                placeholder, ast.TemplateMethodIdentifierNode(placeholder.name))
        elif local_var in self.ast_root.global_identifiers:
            placeholder.parent.replace(placeholder, local_var)
        elif cached_placeholder in local_identifiers:
            placeholder.parent.replace(placeholder, cached_placeholder)
        elif local_var.name in builtin_names:
            placeholder.parent.replace(placeholder,
                                       ast.IdentifierNode(local_var.name))
        elif self.options.cache_resolved_placeholders:
            scope = self.get_parent_scope(placeholder)
            scope.alias_name_set.add(cached_placeholder.name)
            scope.aliased_expression_map[placeholder] = cached_placeholder

            insert_block, insert_marker = self.get_insert_block_and_point(
                placeholder)
            # note: this is sketchy enough that it requires some explanation
            # basically, you need to visit the node for the parent function to
            # get the memo that this value is aliased. unfortunately, the naive
            # case of just calling visit_ast blows up since it tries to double
            # analyze a certain set of nodes. you only really need to analyze
            # that the assignment took place, then you can safely alias the
            # actual function call. definitely sketchy, but it does seem to work
            assign_rph = ast.AssignNode(cached_placeholder,
                                        None,
                                        pos=placeholder.pos)
            cached_placeholder.parent = assign_rph
            #print "optimize scope:", insert_block
            #print "optimize marker:", insert_marker
            insert_block.insert_before(insert_marker, assign_rph)
            self.visit_ast(assign_rph, insert_block)
            assign_rph.right = placeholder
            placeholder.parent.replace(placeholder, cached_placeholder)

    def analyzePlaceholderNode(self, placeholder):
        if self.options.directly_access_defined_variables:
            # when the analyzer finds a ast.PlaceholderNode and generates a
            # function call out of it, i annotate an ast.IdentifierNode with the
            # original placeholder name
            local_var = ast.IdentifierNode(placeholder.name,
                                           pos=placeholder.pos)
            cached_placeholder = ast.IdentifierNode('_rph_%s' % local_var.name,
                                                    pos=placeholder.pos)
            (local_identifiers, partial_local_identifiers, _) = (
                _get_local_identifiers(placeholder))
            attrs = set([ast.IdentifierNode(node.name)
                         for node in self.ast_root.attr_nodes])
            non_local_identifiers = (
                partial_local_identifiers - local_identifiers - attrs)
            if (self.options.static_analysis and
                    local_var in non_local_identifiers):
                self.compiler.error(
                    analyzer.SemanticAnalyzerError(
                        ('Variable %s is not guaranteed to be in scope. '
                         'Define the variable in all branches of the '
                         'conditional or before the conditional.') % local_var),
                    pos=placeholder.pos)
            # print "local_identifiers", local_identifiers
            self._placeholdernode_replacement(
                placeholder, local_var, cached_placeholder, local_identifiers)

    def analyzePlaceholderSubstitutionNode(self, placeholder_substitution):
        self.visit_ast(placeholder_substitution.expression,
                       placeholder_substitution)

#   def alias_expression_in_function(self, function, expression):
#     alias = function.aliased_expression_map.get(expression)
#     if not alias:
#       alias_name = '_%s' % (expression.name)
#       if alias_name in function.alias_name_set:
#         print "duplicate alias_name", alias_name
#         return

#       alias = ast.IdentifierNode(alias_name)
#       function.aliased_expression_map[expression] = alias
#       assign_alias = ast.AssignNode(alias, expression)
#       parent_loop = _get_parent_loop(node)
#       # fixme: check to see if this expression is loop-invariant
#       # must add a test case for this
#       child_node_set = set(node.getChildNodes())
#       #print "child_node_set", child_node_set
#       #print "parent_loop", parent_loop, "parent", node.parent
#       if (parent_loop is not None and
#           not parent_loop.loop_variant_set.intersection(child_node_set)):
#         #print "pull up loop invariant", assign_alias
#         parent_loop.parent.insert_before(parent_loop, assign_alias)
#       else:
#         insert_block, insert_marker = self.get_insert_block_and_point(node)
#         insert_block.insert_before(insert_marker, assign_alias)

#     node.parent.replace(node, alias)

    def analyzeGetAttrNode(self, node):
        if not self.options.alias_invariants:
            return

        # fixme: only handle the trivial case for now
        # simplifies the protocol for making up alias names
        if type(node.expression) != ast.IdentifierNode:
            return

        scope = self.get_parent_scope(node)
        alias = scope.aliased_expression_map.get(node)

        if not alias:
            if node.expression.name[0] != '_':
                alias_format = '_%s_%s'
            else:
                alias_format = '%s_%s'
            alias_name = alias_format % (node.expression.name, node.name)
            if alias_name in scope.alias_name_set:
                print("duplicate alias_name", alias_name)
                print("scope", scope)
                print("scope.alias_name_set", scope.alias_name_set)
                print(" ".join("scope.aliased_expression_map",
                               scope.aliased_expression_map))
                return

            alias = ast.IdentifierNode(alias_name)
            scope.alias_name_set.add(alias_name)
            scope.aliased_expression_map[node] = alias
            assign_alias = ast.AssignNode(alias, node)

            parent_loop = _get_parent_loop(node)
            # fixme: check to see if this expression is loop-invariant
            # must add a test case for this
            child_node_set = set(node.getChildNodes())
            #print "child_node_set", child_node_set
            #print "parent_loop", parent_loop, "parent", node.parent
            if (self.options.inline_hoist_loop_invariant_aliases and
                    parent_loop is not None and
                    not parent_loop.loop_variant_set.intersection(
                        child_node_set)):
                # print "pull up loop invariant", assign_alias
                parent_loop.parent.insert_before(parent_loop, assign_alias)
            else:
                insert_block, insert_marker = self.get_insert_block_and_point(
                    node)
                insert_block.insert_before(insert_marker, assign_alias)

        node.parent.replace(node, alias)

    def analyzeIfNode(self, if_node):
        self.visit_ast(if_node.test_expression, if_node)

        for n in if_node.child_nodes:
            self.visit_ast(n, if_node)

        for n in if_node.else_.child_nodes:
            self.visit_ast(n, if_node.else_)

        parent_scope = self.get_parent_scope(if_node)
        if_scope_vars = if_node.scope.local_identifiers

        # once both branches are optimized, walk the scopes for any variables
        # that are defined in both places. those will be promoted to function
        # scope since it is safe to assume that those will defined fixme: this
        # feels like a bit of hack - but not sure how to do this correctly
        # without reverting to slower performance for almost all calls to
        # resolve_placeholder.
        #
        # it seems like certain optimizations need to be hoisted up to the
        # parent scope. this is particularly the case when you are aliasing
        # common functions that are likely to occur in the parent scope after
        # the conditional block. you *need* to hoist those, or you will have
        # errors when the branch fails. essentially you have to detect and hoist
        # 'branch invariant' optimizations.
        #
        # TODO: we can try to hoist up invariants if they don't depend on the
        # condition. this is somewhat hard to know, so the best way to do so
        # without multiple passes of the optimizer is to hoist only things that
        # were already defined in the parent scope - like _buffer, or things on
        # self.
        if if_node.else_.child_nodes:
            common_local_identifiers = (if_scope_vars
                                        & if_node.else_.scope.local_identifiers)
            # The set of nodes that are not defined in both the if and else
            # branches.
            partial_local_identifiers = (
                (if_scope_vars ^ if_node.else_.scope.local_identifiers)
                | if_node.scope.partial_local_identifiers
                | if_node.else_.scope.partial_local_identifiers)

            common_alias_name_set = (if_node.scope.alias_name_set
                                     & if_node.else_.scope.alias_name_set)

            # Only promote aliased expressions to the parent scope when the
            # alias would be used in both the if and else branches.
            common_aliased_expression_map = _get_common_aliased_expression_map(
                if_node.scope, if_node.else_.scope)

            parent_scope.local_identifiers.update(common_local_identifiers)
            parent_scope.alias_name_set.update(common_alias_name_set)
            parent_scope.aliased_expression_map.update(
                common_aliased_expression_map)
        else:
            partial_local_identifiers = if_scope_vars

        non_parent_scope_identifiers = (
            partial_local_identifiers - parent_scope.local_identifiers)
        parent_scope.partial_local_identifiers.update(
            non_parent_scope_identifiers)
        # Any variable considered dirty in an if or else block should be dirty
        # in the parent scope as well.
        if_dirty_identifiers = if_node.scope.dirty_local_identifiers
        else_dirty_identifiers = if_node.else_.scope.dirty_local_identifiers
        parent_scope.dirty_local_identifiers.update(if_dirty_identifiers)
        parent_scope.dirty_local_identifiers.update(else_dirty_identifiers)

    def analyzeBinOpNode(self, n):
        # if you are trying to use short-circuit behavior, these two
        # optimizations can sabotage correct execution since the rhs may be
        # hoisted above the ast.IfNode and cause it to get executed prior to
        # passing the lhs check.
        should_visit_left = True
        and_or_operator = n.operator in ('and', 'or')
        if and_or_operator:
            self.binop_count += 1
            cache_placeholders = self.options.cache_resolved_placeholders
            cache_udn_expressions = self.options.cache_resolved_udn_expressions
            # If this is the first binop, we can visit the LHS since that must
            # always be executed.
            if self.binop_count == _BINOP_FIRST_PASS:
                self.visit_ast(n.left, n)
                should_visit_left = False
            self.options.cache_resolved_placeholders = False
            self.options.cache_resolved_udn_expressions = False

        if should_visit_left:
            self.visit_ast(n.left, n)
        self.visit_ast(n.right, n)

        if and_or_operator:
            self.binop_count -= 1
            self.options.cache_resolved_placeholders = cache_placeholders
            self.options.cache_resolved_udn_expressions = cache_udn_expressions

    analyzeBinOpExpressionNode = analyzeBinOpNode

    def analyzeUnaryOpNode(self, op_node):
        self.visit_ast(op_node.expression, op_node)

    def analyzeGetUDNNode(self, node):
        if not self.options.prefer_whole_udn_expressions:
            self.visit_ast(node.expression, node)

        # If self._filter_function is created in a macro, make sure we rename
        # it.
        self_node = ast.IdentifierNode('self')
        if node.expression == self_node and node.name == '_filter_function':
            alias = ast.IdentifierNode('_self_private_filter_function',
                                       pos=node.pos)
            node.parent.replace(node, alias)
            return
        # If self.filter_function is created in a macro, make sure we rename it.
        if node.expression == self_node and node.name == 'filter_function':
            alias = ast.IdentifierNode('_self_filter_function', pos=node.pos)
            node.parent.replace(node, alias)
            return

        if self.options.cache_resolved_udn_expressions:
            cached_udn = ast.IdentifierNode(
                _generate_cached_resolved_placeholder(node),
                pos=node.pos)
            (local_identifiers, _, _) = _get_local_identifiers(node)
            if cached_udn in local_identifiers:
                node.parent.replace(node, cached_udn)
            else:
                insert_block, insert_marker = self.get_insert_block_and_point(
                    node)

                # if there is a reassignment in the parent block, don't cache
                # this incase it needs to be re-resolved.
                # #set $text = $text.replace('\r\n', '\n')
                # #set $text = $text.replace('\t', '  ')
                # in this example, if you cache the udn expression text.replace,
                # you have a problem - you won't ever use the new string create
                # by the first call to replace
                for child_node in insert_block.child_nodes:
                    if (isinstance(child_node, ast.AssignNode) and
                            child_node.left == node.expression):
                        return

                scope = self.get_parent_scope(node)
                scope.alias_name_set.add(cached_udn.name)
                scope.aliased_expression_map[node] = cached_udn

                # note: this is sketchy enough that it requires some explanation
                # basically, you need to visit the node for the parent function
                # to get the memo that this value is aliased. unfortunately, the
                # naive case of just calling visit_ast blows up since it tries
                # to double analyze a certain set of nodes. you only really need
                # to analyze that the assignment took place, then you can safely
                # alias the actual function call. definitely sketchy, but it
                # does seem to work
                assign_rph = ast.AssignNode(cached_udn, None, pos=node.pos)
                cached_udn.parent = assign_rph
                insert_block.insert_before(insert_marker, assign_rph)
                self.visit_ast(assign_rph, insert_block)
                assign_rph.right = node
                node.parent.replace(node, cached_udn)
        elif self.options.prefer_whole_udn_expressions:
            self.visit_ast(node.expression, node)

    def analyzeSliceNode(self, pnode):
        self.visit_ast(pnode.expression, pnode)
        self.visit_ast(pnode.slice_expression, pnode)


# a second pass over the optimized tree to hoist invariant aliases to their
# parent blocks
class FinalPassAnalyzer(_BaseAnalyzer):

    def analyzeTemplateNode(self, template):
        self.visit_ast(template.main_function, template)
        for n in template.child_nodes:
            self.visit_ast(n, template)

    def analyzeFunctionNode(self, function):
        for n in function.child_nodes:
            self.visit_ast(n, function)
        if self.options.batch_buffer_writes:
            self.collect_writes(function)

    def analyzeForNode(self, for_node):
        for n in for_node.child_nodes:
            self.visit_ast(n, for_node)

        self.reanalyzeLoopNode(for_node)
        if self.options.batch_buffer_writes:
            self.collect_writes(for_node)

    def analyzeIfNode(self, if_node):
        # depth-first
        for n in if_node.child_nodes:
            self.visit_ast(n, if_node)

        for n in if_node.else_.child_nodes:
            self.visit_ast(n, if_node.else_)

        self.reanalyzeConditionalNode(if_node)
        self.reanalyzeConditionalNode(if_node.else_)
        if self.options.batch_buffer_writes:
            self.collect_writes(if_node)
            self.collect_writes(if_node.else_)

    def analyzeBufferWrite(self, buffer_write):
        """Perform ast.BufferWrite optimizations.

    Do this in the final pass optimizer to make sure that the
    optimization is handled after caching placeholders.
    """
        self.visit_ast(buffer_write.expression, buffer_write)
        # All filterning is done before writing to the buffer. If the function
        # output needed filtering then it would be wrapped in a ast.FilterNode.
        if isinstance(buffer_write.expression, ast.CallFunctionNode):
            buffer_write.expression.sanitization_state = (
                ast.SanitizedState.OUTPUTTED_IMMEDIATELY)

    def hoist(self, parent_node, parent_block, insertion_point, alias_node,
              assign_alias_node):

        # prune the implementation in the nested block
        # print "prune", alias_node
        # print " ".join("parent_block aliases",
        #                parent_block.scope.aliased_expression_map)
        parent_node.remove(assign_alias_node)
        # if we've already hoisted an assignment, don't do it again
        if alias_node not in parent_block.scope.hoisted_aliases:
            # prune the original implementation in the current block and
            # reinsert the alias before it's first potential usage if it
            # is needed earlier in the execution path.
            # when a variable aliased in both the if and
            # else blocks is promoted to the parent scope
            # the implementation isn't actually hoisted (should it be?)
            # inline with the ast.IfNode optimization so we need to check if the
            # node is already here
            if assign_alias_node in parent_block.child_nodes:
                current_pos = parent_block.child_nodes.index(assign_alias_node)
                # an else node's parent is the ast.IfNode, which is the relevant
                # node when searching for the insertion point
                needed_pos = parent_block.child_nodes.index(insertion_point)
                if needed_pos < current_pos:
                    parent_block.child_nodes.remove(assign_alias_node)
                    if isinstance(parent_node, ast.ElseNode):
                        parent_block.insert_before(parent_node.parent,
                                                   assign_alias_node)
                    else:
                        parent_block.insert_before(parent_node,
                                                   assign_alias_node)
                    # print "insert_before", alias_node
            else:
                # still need to insert the alias
                parent_block.insert_before(insertion_point, assign_alias_node)
            parent_block.scope.hoisted_aliases.append(alias_node)

        # NOTE: once we hoist an expression, we need to make sure that we no
        # longer use this for dependencies in the current scope
        del parent_node.scope.aliased_expression_map[alias_node]
        parent_node.scope.alias_name_set.remove(assign_alias_node.left.name)
        # FIXME: this is probably an indication of a bug or unnecessary
        # difference between the caching of placeholders and filter expressions
        if not isinstance(alias_node, ast.FilterNode):
            parent_node.scope.local_identifiers.remove(assign_alias_node.left)

    def make_write_node(self, write_list):
        """Convert a list of expressions to be written into a single
        ast.ASTNode. This will return either be a ast.BufferWrite node, a
        ast.BufferExtend node or None if write_list is empty.
        """
        if not write_list:
            return None
        # If there is only a single node, avoid the overhead of
        # constructing a tuple.
        if len(write_list) == 1:
            return ast.BufferWrite(write_list[0])
        tuple_node = ast.TupleLiteralNode()
        for exp in write_list:
            tuple_node.append(exp)
        return ast.BufferExtend(tuple_node)

    def collect_writes(self, node):
        """Look at all the writes that are children of the node and move them to
        the latest place they can be safely placed. This means that it either
        goes before an (if/function/loop) or at the end of the list of nodes.
        Once they are all together, they can be batched into a buffer.extend
        operation. This avoids list resize operations.
        """

        passable_nodes = (ast.AssignNode, ast.CacheNode)

        # ast.ASTNode.insert_before is broken in a way that if there are
        # duplicate nodes, inserts always occur before the first one. To
        # deal with this, we construct a new ast.NodeList rather than
        # modifying the original one.
        old_node_list = node.child_nodes
        node.child_nodes = ast.NodeList()

        write_list = []
        for n in old_node_list:
            if isinstance(n, ast.BufferWrite):
                write_list.append(n.expression)
            elif isinstance(n, ast.BufferExtend):
                write_list.extend(n.expression.child_nodes)
            elif not isinstance(n, passable_nodes):
                write_node = self.make_write_node(write_list)
                if write_node:
                    node.append(write_node)
                write_list = []
                node.append(n)
            else:
                node.append(n)
        write_node = self.make_write_node(write_list)
        if write_node:
            node.append(write_node)


template_function_re = re.compile('^[^#]*#(def|block)\s+(\w+)')
extends_re = re.compile('^#extends\s+([\.\w]+)')
template_extensions = ('.spt', '.tmpl')


def _extend_to_real_path(base_dir, ex_path):
    for ext in template_extensions:
        rpath = os.path.join(base_dir, ex_path) + ext
        if os.path.exists(rpath):
            return rpath
    raise Exception(
        'could not find .spt or .tmpl file for %s during dependency check' %
        ex_path)


# scan an spt file for template functions it will output
def get_template_functions(base_dir, path):
    template_function_names = set()
    path = _extend_to_real_path(base_dir, path)
    if not path:
        return template_function_names
    f = open(path)
    for line in f:
        match = template_function_re.match(line)
        if match:
            template_function_names.add(match.group(2))
            continue
        match = extends_re.match(line)
        if match:
            extend_name = match.group(1)
            extend_path = extend_name.replace('.', '/')
            template_function_names.update(get_template_functions(base_dir,
                                                                  extend_path))
    f.close()
    return template_function_names



# Yapps 2.0 Runtime
#
# This module is needed to run generated parsers.

import logging
import re
import io
import sys

class SyntaxError(Exception):
  """When we run into an unexpected token, this is the exception to use"""
  def __init__(self, pos=-1, msg="Bad Token"):
    Exception.__init__(self)
    self.pos = pos
    self.msg = msg
  def __repr__(self):
    if self.pos < 0:
      return "#<syntax-error>"
    else:
      return "SyntaxError[@ char %s: %s]" % (repr(self.pos), self.msg)


class NoMoreTokens(Exception):
  """Another exception object, for when we run out of tokens"""
  pass

class FatalParseError(Exception):
  """We have to fail-stop for one reason or another."""

class Scanner(object):
  def __init__(self, patterns, ignore, input_):
    """Patterns is [(terminal,regex)...]
    Ignore is [terminal,...];
    Input is a string"""
    self.tokens = []
    self.restrictions = []
    self.input = input_
    self.pos = 0
    self.ignore = ignore
    # The stored patterns are a pair (compiled regex,source
    # regex).  If the patterns variable passed in to the
    # constructor is None, we assume that the class already has a
    # proper .patterns list constructed
    if patterns is not None:
      self.patterns = []
      for k, r in patterns:
        self.patterns.append( (k, re.compile(r)) )

  def file_position(self, token_index):
    """Returns the in-file start position of the token at the given index.

    The token must be one of the already read tokens,
    or the next one we are about to read.  Otherwise,
    this raises IndexError."""
    if token_index < len(self.tokens):
      return self.tokens[token_index][0]
    elif token_index == len(self.tokens):
      return self.pos
    else:
      raise IndexError(
          "Can't determine file position of future token %d." % token_index)

  def token(self, i, restrict=0):
    """Get the i'th token, and if i is one past the end, then scan
    for another token; restrict is a list of tokens that
    are allowed, or 0 for any token."""
    if i == len(self.tokens):
      self.scan(restrict)
    if i < len(self.tokens):
      # Make sure the restriction is more restricted
      if restrict and self.restrictions[i]:
        for r in restrict:
          if r not in self.restrictions[i]:
            raise NotImplementedError(
              "Unimplemented: restriction set changed", r, self.restrictions[i])
        return self.tokens[i]
      elif not restrict and not self.restrictions[i]:
        return self.tokens[i]
    raise NoMoreTokens(i, len(self.tokens), self.tokens[i], restrict, self.restrictions[i], self.tokens)

  def __repr__(self):
    """Print the last 10 tokens that have been scanned in"""
    output = ''
    for t in self.tokens[-10:]:
      output = '%s\n  (@%s)  %s  =  %s' % (output,t[0],t[2],repr(t[3]))
    return output

  def scan(self, restrict):
    """Should scan another token and add it to the list, self.tokens,
    and add the restriction to self.restrictions"""
    # Keep looking for a token, ignoring any in self.ignore
    while 1:
      # Search the patterns for the longest match, with earlier
      # tokens in the list having preference
      best_match = -1
      best_pat = '(error)'
      for p, regexp in self.patterns:
        # First check to see if we're ignoring this token
        if restrict and p not in restrict and p not in self.ignore:
          continue
        m = regexp.match(self.input, self.pos)
        if m and len(m.group(0)) > best_match:
          # We got a match that's better than the previous one
          best_pat = p
          best_match = len(m.group(0))

      # If we didn't find anything, raise an error
      if best_pat == '(error)' and best_match < 0:
        msg = "Bad Token"
        if restrict:
          msg = "Trying to find one of "+restrict+", "
        raise SyntaxError(self.pos, msg)

      # If we found something that isn't to be ignored, return it
      if best_pat not in self.ignore:
        # Create a token with this data
        token = (self.pos, self.pos+best_match, best_pat,
                 self.input[self.pos:self.pos+best_match])
        self.pos = self.pos + best_match
        # Only add this token if it's not in the list
        # (to prevent looping)
        if not self.tokens or token != self.tokens[-1]:
          self.tokens.append(token)
          self.restrictions.append(restrict)
          return
      else:
        # This token should be ignored ..
        self.pos = self.pos + best_match


class Parser(object):
  def __init__(self, scanner):
    self._scanner = scanner
    self._pos = 0

  def _peek(self, *types):
    """Returns the token type for lookahead; if there are any args
    then the list of args is the set of token types to allow"""
    tok = self._scanner.token(self._pos, types)
    return tok[2]

  def _scan(self, type):
    """Returns the matched text, and moves to the next token"""
    tok = self._scanner.token(self._pos, [type])
    if tok[2] != type:
      raise SyntaxError(tok[0], 'Trying to find %s (%s)' % (type, tok[0]))
    self._pos = 1+self._pos
    return tok[3]

  @property
  def file_position(self):
    return self._scanner.file_position(self._pos)


def format_error(input_, err, scanner):
  """This is a really dumb long function to print error messages nicely."""
  error_message = io.StringIO()
  p = err.pos
  print("error position", p, file=error_message)
  # Figure out the line number
  line = input_[:p].count("\n")
  print(err.msg, "on line", repr(line+1) + ":", file=error_message)
  # Now try printing part of the line
  text = input_[max(p-80, 0):p+80]
  p = p - max(p-80, 0)

  # Strip to the left
  i = text[:p].rfind("\n")
  j = text[:p].rfind("\n")
  if i < 0 or (0 <= j < i): i = j
  if 0 <= i < p:
    p = p - i - 1
  text = text[i+1:]

  # Strip to the right
  i = text.find('\n', end=p)
  i = text.find('\r', end=p)
  if i < 0 or (0 <= j < i):
    i = j
  if i >= 0:
    text = text[:i]

  # Now shorten the text
  while len(text) > 70 and p > 60:
    # Cut off 10 chars
    text = "..." + text[10:]
    p = p - 7

  # Now print the string, along with an indicator
  print('> ', text.replace('\t', ' ').encode(sys.getdefaultencoding()), file=error_message)
  print('> ', ' '*p + '^', file=error_message)
  print('List of nearby tokens:', scanner, file=error_message)
  return error_message.getvalue()


def wrap_error_reporter(parser, rule):
  try:
    return getattr(parser, rule)()
  except SyntaxError as e:
    logging.exception('syntax error')
    input_ = parser._scanner.input
    try:
      error_msg = format_error(input_, e, parser._scanner)
    except ImportError:
      error_msg = 'Syntax Error %s on line\n' % (e.msg, 1 + input_[:e.pos].count("\n"))
  except NoMoreTokens as e:
    error_msg = 'Could not complete parsing; stopped around here:\n%s\n%s' % (parser._scanner, e)
  raise FatalParseError(error_msg)



# Copyright 2014 The Spitfire Authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import unittest

from spitfire.compiler import analyzer
from spitfire.compiler import ast
from spitfire.compiler import compiler
from spitfire.compiler import options
from spitfire.compiler import util
from spitfire.compiler import walker
from spitfire import test_util


class BaseTest(unittest.TestCase):

    def __init__(self, *args):
        unittest.TestCase.__init__(self, *args)
        self.analyzer_options = options.default_options
        self.analyzer_options.update(cache_resolved_placeholders=True,
                                     enable_warnings=True,
                                     warnings_as_errors=True)

    def setUp(self):
        self.compiler = compiler.Compiler(
            analyzer_options=self.analyzer_options,
            xspt_mode=False,
            compiler_stack_traces=True)

    def _get_analyzer(self, ast_root):
        semantic_analyzer = analyzer.SemanticAnalyzer(
            'TestTemplate', ast_root, self.compiler.analyzer_options,
            self.compiler)
        semantic_analyzer.get_ast = test_util.RecordedFunction(
            semantic_analyzer.get_ast)
        return semantic_analyzer

    def _build_function_template(self):
        """ Build a simple template with a function.

        file: TestTemplate
        #def test_function
        #end def
        """
        ast_root = ast.TemplateNode('TestTemplate')
        def_node = ast.DefNode('test_function')
        ast_root.append(def_node)
        return (ast_root, def_node)

    def _build_if_template(self, condition=None):
        """ Build a simple template with a function and an if statement.

        file: TestTemplate
        #def test_function
          #if True
          #end if
        #end def
        """
        ast_root, def_node = self._build_function_template()
        condition_node = condition or ast.LiteralNode(True)
        if_node = ast.IfNode(condition_node)
        def_node.append(if_node)
        return (ast_root, def_node, if_node)

    def _compile(self, template_content):
        template_node = util.parse_template(template_content)
        template_node.source_path = 'test_template.spt'
        return template_node


class TestEmptyIfBlockError(BaseTest):

    def test_empty_if_fails(self):
        self.ast_description = """
        file: TestTemplate
        #def test_function
          #if True
          #end if
        #end def
        """

        ast_root, def_node, if_node = self._build_if_template()

        semantic_analyzer = self._get_analyzer(ast_root)

        self.assertRaises(analyzer.SemanticAnalyzerError,
                          semantic_analyzer.get_ast)

    def test_optional_whitespace_if_fails(self):
        self.ast_description = """
        file: TestTemplate
        #def test_function
          #if True

          #end if
        #end def
        """

        ast_root, def_node, if_node = self._build_if_template()
        if_node.append(ast.OptionalWhitespaceNode(' '))

        semantic_analyzer = self._get_analyzer(ast_root)

        self.assertRaises(analyzer.SemanticAnalyzerError,
                          semantic_analyzer.get_ast)

    def test_comment_if_fails(self):
        self.ast_description = """
        file: TestTemplate
        #def test_function
          #if True
            ## This is a comment.
          #end if
        #end def
        """

        ast_root, def_node, if_node = self._build_if_template()
        if_node.append(ast.CommentNode(' This is a comment.'))

        semantic_analyzer = self._get_analyzer(ast_root)

        self.assertRaises(analyzer.SemanticAnalyzerError,
                          semantic_analyzer.get_ast)

    def test_empty_if_full_else_fails(self):
        self.ast_description = """
        file: TestTemplate
        #def test_function
          #if True
          #else
            #set $foo = true
          #end if
        #end def
        """

        ast_root, def_node, if_node = self._build_if_template()
        assign_node = ast.AssignNode(
            ast.IdentifierNode('foo'), ast.LiteralNode(True))
        if_node.else_.append(assign_node)

        semantic_analyzer = self._get_analyzer(ast_root)

        self.assertRaises(analyzer.SemanticAnalyzerError,
                          semantic_analyzer.get_ast)

    def test_empty_elif_fails(self):
        self.ast_description = """
        file: TestTemplate
        #def test_function
          #if True
            #set $foo = True
          #elif False
          #end if
        #end def
        """

        ast_root, def_node, if_node = self._build_if_template()
        assign_node = ast.AssignNode(
            ast.IdentifierNode('foo'), ast.LiteralNode(True))
        if_node.append(assign_node)
        elif_node = ast.IfNode(ast.LiteralNode(False))
        if_node.else_.append(elif_node)
        semantic_analyzer = self._get_analyzer(ast_root)

        self.assertRaises(analyzer.SemanticAnalyzerError,
                          semantic_analyzer.get_ast)

    def test_non_empty_if_ok(self):
        self.ast_description = """
        file: TestTemplate
        #def test_function
          #if True
            #set $foo = True
          #end if
        #end def
        """

        ast_root, def_node, if_node = self._build_if_template()
        assign_node = ast.AssignNode(
            ast.IdentifierNode('foo'), ast.LiteralNode(True))
        if_node.append(assign_node)

        semantic_analyzer = self._get_analyzer(ast_root)

        try:
            semantic_analyzer.get_ast()
        except analyzer.SemanticAnalyzerError:
            self.fail('get_ast raised SemanticAnalyzerError unexpectedly.')


class TestEmptyForBlockError(BaseTest):

    def _build_for_template(self):
        """ Build a simple template with a function and a for loop.

        file: TestTemplate
        #def test_function
          #for $i in []
          #end for
        #end def
        """
        ast_root, def_node = self._build_function_template()
        target_list = ast.TargetListNode()
        target_list.append(ast.PlaceholderNode('foo'))
        expression_list = ast.ExpressionListNode()
        expression_list.append(ast.LiteralNode([]))
        for_node = ast.ForNode(target_list=target_list,
                               expression_list=expression_list)
        def_node.append(for_node)
        return (ast_root, def_node, for_node)

    def test_empty_for_fails(self):
        self.ast_description = """
        file: TestTemplate
        #def test_function
          #for $i in []
          #end for
        #end def
        """

        ast_root, def_node, for_node = self._build_for_template()

        semantic_analyzer = self._get_analyzer(ast_root)

        self.assertRaises(analyzer.SemanticAnalyzerError,
                          semantic_analyzer.get_ast)

    def test_optional_whitespace_for_fails(self):
        self.ast_description = """
        file: TestTemplate
        #def test_function
          #for $i in []

          #end for
        #end def
        """

        ast_root, def_node, for_node = self._build_for_template()
        for_node.append(ast.OptionalWhitespaceNode(' '))

        semantic_analyzer = self._get_analyzer(ast_root)

        self.assertRaises(analyzer.SemanticAnalyzerError,
                          semantic_analyzer.get_ast)

    def test_comment_for_fails(self):
        self.ast_description = """
        file: TestTemplate
        #def test_function
          #for $i in []
            ## This is a comment.
          #end for
        #end def
        """

        ast_root, def_node, for_node = self._build_for_template()
        for_node.append(ast.CommentNode(' This is a comment.'))

        semantic_analyzer = self._get_analyzer(ast_root)

        self.assertRaises(analyzer.SemanticAnalyzerError,
                          semantic_analyzer.get_ast)

    def test_non_empty_for_ok(self):
        self.ast_description = """
        file: TestTemplate
        #def test_function
          #for $i in []
             #set $foo = True
          #end for
        #end def
        """

        ast_root, def_node, for_node = self._build_for_template()
        assign_node = ast.AssignNode(
            ast.IdentifierNode('foo'), ast.LiteralNode(True))
        for_node.append(assign_node)

        semantic_analyzer = self._get_analyzer(ast_root)

        try:
            semantic_analyzer.get_ast()
        except analyzer.SemanticAnalyzerError:
            self.fail('get_ast raised SemanticAnalyzerError unexpectedly.')


class TestGlobalScopeLibraryError(BaseTest):

    def _build_function_template_library(self):
        """ Build a simple library template with a function.

        file: TestTemplate
        #implements library
        #def test_function
        #end def
        """
        ast_root = ast.TemplateNode('TestTemplate')
        implements_node = ast.ImplementsNode('library')
        ast_root.append(implements_node)
        def_node = ast.DefNode('test_function')
        ast_root.append(def_node)
        return (ast_root, def_node)

    def test_library_ok(self):
        self.ast_description = """
        file: TestTemplate
        #implements library
        #def test_function
        #end def
        """

        ast_root, def_node = self._build_function_template_library()

        semantic_analyzer = self._get_analyzer(ast_root)

        try:
            semantic_analyzer.get_ast()
        except analyzer.SemanticAnalyzerError:
            self.fail('get_ast raised SemanticAnalyzerError unexpectedly.')

    def test_set_error(self):
        self.ast_description = """
        file: TestTemplate
        #implements library
        #set $foo = True
        #def test_function
        #end def
        """

        ast_root, def_node = self._build_function_template_library()
        assign_node = ast.AssignNode(
            ast.IdentifierNode('foo'), ast.LiteralNode(True))
        ast_root.insert_before(def_node, assign_node)

        semantic_analyzer = self._get_analyzer(ast_root)

        self.assertRaises(analyzer.SemanticAnalyzerError,
                          semantic_analyzer.get_ast)

    def test_attr_error(self):
        self.ast_description = """
        file: TestTemplate
        #implements library
        #attr $foo = True
        #def test_function
        #end def
        """

        ast_root, def_node = self._build_function_template_library()
        attr_node = ast.AttributeNode('foo', default=ast.LiteralNode(True))
        ast_root.insert_before(def_node, attr_node)

        semantic_analyzer = self._get_analyzer(ast_root)

        self.assertRaises(analyzer.SemanticAnalyzerError,
                          semantic_analyzer.get_ast)

    def test_global_ok(self):
        self.ast_description = """
        file: TestTemplate
        #implements library
        #global $foo
        #def test_function
        #end def
        """

        ast_root, def_node = self._build_function_template_library()
        global_node = ast.GlobalNode('foo')
        ast_root.insert_before(def_node, global_node)

        semantic_analyzer = self._get_analyzer(ast_root)

        try:
            semantic_analyzer.get_ast()
        except analyzer.SemanticAnalyzerError:
            self.fail('get_ast raised SemanticAnalyzerError unexpectedly.')


class TestAssignSlice(BaseTest):

    def test_slice_non_identifier_error(self):
        self.ast_description = """
        file: TestTemplate
        #def test_function
          #set 1[1] = 1
        #end def
        """

        ast_root, def_node = self._build_function_template()
        assign_node = ast.AssignNode(
            ast.SliceNode(
                ast.LiteralNode(1), ast.LiteralNode(1)), ast.LiteralNode(1))
        def_node.append(assign_node)

        semantic_analyzer = self._get_analyzer(ast_root)

        self.assertRaises(analyzer.SemanticAnalyzerError,
                          semantic_analyzer.get_ast)

    def test_slice_identifier_ok(self):
        self.ast_description = """
        file: TestTemplate
        #def test_function
          #set $foo[1] = 1
        #end def
        """

        ast_root, def_node = self._build_function_template()
        assign_node = ast.AssignNode(
            ast.SliceNode(
                ast.IdentifierNode('foo'), ast.LiteralNode(1)),
            ast.LiteralNode(1))
        def_node.append(assign_node)

        semantic_analyzer = self._get_analyzer(ast_root)

        try:
            semantic_analyzer.get_ast()
        except analyzer.SemanticAnalyzerError:
            self.fail('get_ast raised SemanticAnalyzerError unexpectedly.')


class TestSanitizedFunction(BaseTest):

    def setUp(self):
        self.analyzer_options = options.default_options
        self.analyzer_options.update(cache_resolved_placeholders=True,
                                     enable_warnings=True,
                                     warnings_as_errors=True,
                                     baked_mode=True,
                                     generate_unicode=False)
        self.compiler = compiler.Compiler(
            analyzer_options=self.analyzer_options,
            xspt_mode=False,
            compiler_stack_traces=True)
        self.compiler.new_registry_format = True
        self.compiler.function_name_registry['reg_f'] = ('a.reg_f',
                                                         ['skip_filter'])

    def test_template_method_direct(self):
        code = """
        #def foo
          Hello
        #end def

        #def bar
          $foo()
        #end def
        """

        template = self._compile(code)
        semantic_analyzer = self._get_analyzer(template)
        analyzed_ast = semantic_analyzer.get_ast()

        def pred(node):
            return (type(node) == ast.CallFunctionNode and
                    type(node.expression) == ast.PlaceholderNode and
                    node.expression.name == 'foo')

        foo_call = walker.find_node(analyzed_ast, pred)
        if not foo_call:
            self.fail('Expected foo() in ast')
        self.assertEqual(foo_call.sanitization_state,
                         ast.SanitizedState.SANITIZED_STRING)

    def test_library_function_direct(self):
        code = """
        #from module import library my_lib

        #def bar
          $my_lib.foo()
        #end def
        """

        template = self._compile(code)
        semantic_analyzer = self._get_analyzer(template)
        analyzed_ast = semantic_analyzer.get_ast()

        def pred(node):
            return (type(node) == ast.CallFunctionNode and
                    type(node.expression) == ast.IdentifierNode and
                    node.expression.name == 'my_lib.foo')

        foo_call = walker.find_node(analyzed_ast, pred)
        if not foo_call:
            self.fail('Expected my_lib.foo() in ast')
        self.assertEqual(foo_call.sanitization_state,
                         ast.SanitizedState.SANITIZED_STRING)

    def test_library_function_registry_yes(self):
        code = """
        #def bar
          $reg_f()
        #end def
        """

        template = self._compile(code)
        semantic_analyzer = self._get_analyzer(template)
        analyzed_ast = semantic_analyzer.get_ast()

        def pred(node):
            return (type(node) == ast.CallFunctionNode and
                    type(node.expression) == ast.PlaceholderNode and
                    node.expression.name == 'reg_f')

        foo_call = walker.find_node(analyzed_ast, pred)
        if not foo_call:
            self.fail('Expected reg_f() in ast')
        self.assertEqual(foo_call.sanitization_state,
                         ast.SanitizedState.SANITIZED)

    def test_external_function_maybe(self):
        code = """
        #from module import my_lib

        #def bar
          $my_lib.foo()
        #end def
        """

        template = self._compile(code)
        semantic_analyzer = self._get_analyzer(template)
        analyzed_ast = semantic_analyzer.get_ast()

        def pred(node):
            return (type(node) == ast.CallFunctionNode and
                    type(node.expression) == ast.GetUDNNode and
                    type(node.expression.expression) == ast.PlaceholderNode and
                    node.expression.expression.name == 'my_lib' and
                    node.expression.name == 'foo')

        foo_call = walker.find_node(analyzed_ast, pred)
        if not foo_call:
            self.fail('Expected my_libfoo() in ast')
        self.assertEqual(foo_call.sanitization_state,
                         ast.SanitizedState.UNKNOWN)


class TestNoRaw(BaseTest):

    def setUp(self):
        self.analyzer_options = options.default_options
        self.analyzer_options.update(enable_warnings=True,
                                     warnings_as_errors=True,
                                     no_raw=True)
        self.compiler = compiler.Compiler(
            analyzer_options=self.analyzer_options,
            xspt_mode=False,
            compiler_stack_traces=True)

    def test_error_with_raw(self):
        code = """
        #def foo
          #set $a = "a"
          ${a|raw}
        #end def
        """

        template = self._compile(code)
        semantic_analyzer = self._get_analyzer(template)
        self.assertRaises(analyzer.SemanticAnalyzerError,
                          semantic_analyzer.get_ast)

    def test_allow_raw_no_error(self):
        code = """
        #allow_raw
        #def foo
          #set $a = "a"
          ${a|raw}
        #end def
        """

        template = self._compile(code)
        semantic_analyzer = self._get_analyzer(template)
        try:
            semantic_analyzer.get_ast()
        except analyzer.SemanticAnalyzerError:
            self.fail('get_ast raised an error unexpectedly.')

    def test_allow_raw_macro_no_error(self):
        code = """
        #allow_raw
        #global $a
        #def foo
          #i18n()#
            ${a|raw}
          #end i18n#
        #end def
        """

        template = self._compile(code)
        semantic_analyzer = self._get_analyzer(template)
        try:
            semantic_analyzer.get_ast()
        except analyzer.SemanticAnalyzerError:
            self.fail('get_ast raised an error unexpectedly.')

    def test_allow_raw_no_raw_error(self):
        code = """
        #allow_raw
        #def foo
        #end def
        """

        template = self._compile(code)
        semantic_analyzer = self._get_analyzer(template)
        self.assertRaises(analyzer.SemanticAnalyzerError,
                          semantic_analyzer.get_ast)


if __name__ == '__main__':
    unittest.main()



# Copyright 2007 The Spitfire Authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import os.path
import platform
from setuptools import Extension
from setuptools import setup

import spitfire

NAME = 'spitfire'

DESCRIPTION = 'text-to-python template language'

VERSION = spitfire.__version__

AUTHOR = spitfire.__author__

AUTHOR_EMAIL = spitfire.__author_email__

LICENSE = spitfire.__license__

PLATFORMS = ['Posix', 'MacOS X', 'Windows']

CLASSIFIERS = ['Development Status :: 3 - Alpha',
               'Intended Audience :: Developers',
               'License :: OSI Approved :: BSD License',
               'Programming Language :: Python',
               'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
               'Topic :: Software Development :: Code Generators',
               'Topic :: Text Processing']

PACKAGES = ['spitfire',
            'spitfire.compiler',
            'spitfire.compiler.macros',
            'spitfire.runtime']

PY_MODULES = ['third_party.yapps2.yappsrt']

SCRIPTS = ['scripts/crunner.py', 'scripts/spitfire-compile']

EXT_MODULES = None

setup(name=NAME,
      description=DESCRIPTION,
      version=VERSION,
      author=AUTHOR,
      author_email=AUTHOR_EMAIL,
      license=LICENSE,
      platforms=PLATFORMS,
      classifiers=CLASSIFIERS,
      packages=PACKAGES,
      py_modules=PY_MODULES,
      scripts=SCRIPTS,
      ext_modules=EXT_MODULES)



# Copyright 2007 The Spitfire Authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.


class __UnresolvedPlaceholder(object):
    pass


UnresolvedPlaceholder = __UnresolvedPlaceholder()


class PlaceholderError(KeyError):
    pass


class UDNResolveError(Exception):
    pass


# the idea is to have something that is always like None, but explodes when
# you try to use it as a string. this means that you can resolve placeholders
# and evaluate them in complex conditional expressions, allowing them to be
# hoisted, and still protect conditional access to the values
# it could also be that you might try to call the result - in that case, blow
# and exception as well.
class UndefinedPlaceholder(object):

    def __init__(self, name, search_list):
        # Uses mangled names for internal state in case these attributes
        # happen to match the object it's replacing.  eg, if the user_object
        # was an UndefinedPlaceholder in this statement:  $user_object.name
        self.__name = name
        self.__search_list = search_list

    def __bool__(self):
        return False

    def __str__(self):
        raise PlaceholderError(self.__name, self.get_placeholder_names())

    def __call__(self, *pargs, **kargs):
        raise PlaceholderError(self.__name, self.get_placeholder_names())

    def get_placeholder_names(self):
        return [_get_available_placeholders(scope)
                for scope in self.__search_list]


class UndefinedAttribute(UndefinedPlaceholder):
    pass


def _get_available_placeholders(scope):
    if isinstance(scope, dict):
        return list(scope.keys())
    else:
        return [a
                for a in dir(scope)
                if not (a.startswith('__') and a.endswith('__'))]


def import_module_symbol(name):
    name_parts = name.split('.')
    module_name = '.'.join(name_parts[:-1])
    symbol_name = name_parts[-1]
    module = __import__(module_name, globals(), locals(), [symbol_name])
    try:
        symbol = getattr(module, symbol_name)
    except AttributeError as e:
        raise ImportError("can't import %s" % name)
    return symbol


# map template function names to python function names
# inject them into a module so they run as globals
def register_functions(module, template_function_map):
    for t_name, f_name in template_function_map.items():
        f_func = import_module_symbol(f_name)
        setattr(module, t_name, f_func)


# decorate a function object so the value will be retrieved once and then
# cached in the template forever.
def cache_forever(function):
    function.cache_forever = True
    return function


# decorate a function object so its result is not cached in module globals
def never_cache(function):
    function.never_cache = True
    return function



# Copyright 2014 The Spitfire Authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import copy
import gc
import os.path
import sys

from spitfire.compiler import analyzer
from spitfire.compiler import codegen
from spitfire.compiler import optimizer
from spitfire.compiler import options
from spitfire.compiler import util


class CompilerError(Exception):
    pass


class Warning(Exception):
    pass


class Compiler(object):
    setting_names = [
        'baked_mode',
        'base_extends_package',
        'base_template_full_import_path',
        'debug_flags',
        'compiler_stack_traces',
        'default_to_strict_resolution',
        'double_assign_error',
        'enable_filters',
        'extract_message_catalogue',
        'fail_library_searchlist_access',
        'function_registry_file',
        'ignore_optional_whitespace',
        'include_sourcemap',
        'include_path',
        'locale',
        'message_catalogue_file',
        'no_raw',
        'normalize_whitespace',
        'optimizer_flags',
        'optimizer_level',
        'output_directory',
        'skip_import_udn_resolution',
        'static_analysis',
        'strict_global_check',
        'tune_gc',
        'enable_warnings',
        'warnings_as_errors',
        'xspt_mode',
    ]

    @classmethod
    def args_from_optparse(cls, options):
        settings = {}
        for name in cls.setting_names:
            if hasattr(options, name):
                settings[name] = getattr(options, name)
        return settings

    # settings - arbitrary dictionary of values, probably from the command line
    def __init__(self, **kargs):
        # record transient state of the compiler
        self.src_filename = None
        self.src_text = None
        self.src_line_map = []
        self.output_directory = ''
        self.xspt_mode = False
        self.write_file = False
        self.analyzer_options = None
        self.include_sourcemap = False
        # This context can be used by macros to handle state when
        # compiling a file. Each macro should use its own namespace to
        # avoid conflicts. This is not enforced in any way. The context is
        # reset after compiling a single file. A simple example is
        # counting the number of times a macro has run.
        self.macro_context = {}

        self.optimizer_level = 0
        self.optimizer_flags = []
        self.debug_flags = []
        self.compiler_stack_traces = False
        self.ignore_optional_whitespace = False
        self.normalize_whitespace = False
        self.fail_library_searchlist_access = False
        self.skip_import_udn_resolution = False
        self.default_to_strict_resolution = False
        self.static_analysis = False
        self.strict_global_check = False
        self.double_assign_error = False
        self.enable_warnings = False
        self.warnings_as_errors = False
        self.baked_mode = False
        self.no_raw = False

        self.base_extends_package = None
        self.base_template_full_import_path = None
        self.message_catalogue = None
        self.message_catalogue_file = None
        self.extract_message_catalogue = False
        self.locale = None
        self.include_path = '.'
        self.enable_filters = True
        self.tune_gc = False

        # the function registry is for optimized access to 'first-class'
        # functions things that get accessed all the time that should be speedy
        self.function_registry_file = None
        self.function_name_registry = {}
        self.macro_registry = {}

        self._parse_tree = None
        self._analyzed_tree = None
        self._optimized_tree = None
        self._hoisted_tree = None
        self._source_code = None

        for key, value in kargs.items():
            setattr(self, key, value)

        if self.analyzer_options is None:
            self.analyzer_options = options.optimizer_map[self.optimizer_level]
            if self.base_template_full_import_path:
                self.analyzer_options.base_template_full_import_path = (
                    self.base_template_full_import_path)
            self.analyzer_options.ignore_optional_whitespace = (
                self.ignore_optional_whitespace)
            self.analyzer_options.normalize_whitespace = (
                self.normalize_whitespace)
            self.analyzer_options.fail_library_searchlist_access = (
                self.fail_library_searchlist_access)
            self.analyzer_options.skip_import_udn_resolution = (
                self.skip_import_udn_resolution)
            self.analyzer_options.static_analysis = self.static_analysis
            self.analyzer_options.strict_global_check = self.strict_global_check
            self.analyzer_options.double_assign_error = self.double_assign_error
            self.analyzer_options.default_to_strict_resolution = (
                self.default_to_strict_resolution)
            self.analyzer_options.include_sourcemap = self.include_sourcemap
            self.analyzer_options.baked_mode = self.baked_mode
            self.analyzer_options.no_raw = self.no_raw

        # slightly crappy code to support turning flags on and off from the
        # command line - probably should go in analyzer options?
        for flag_name in self.optimizer_flags:
            flag_name = flag_name.lower().replace('-', '_')
            flag_value = True
            if flag_name.startswith('no_'):
                flag_name = flag_name[3:]
                flag_value = False
            if type(getattr(self.analyzer_options, flag_name, None)) == bool:
                setattr(self.analyzer_options, flag_name, flag_value)
            else:
                logging.warning('unknown optimizer flag: %s', flag_name)

        if self.function_registry_file:
            self.new_registry_format, self.function_name_registry = (
                util.read_function_registry(self.function_registry_file))

        # register macros before the first pass by any SemanticAnalyzer this is
        # just a default to give an example - it's not totally functional fixme:
        # nasty time to import - but does break weird cycle
        import spitfire.compiler.macros.i18n
        self.register_macro('macro_i18n',
                            spitfire.compiler.macros.i18n.macro_i18n)
        self.register_macro('macro_function_i18n',
                            spitfire.compiler.macros.i18n.macro_function_i18n)

    # take an AST and generate code from it - this will run the analysis phase
    # this doesn't have the same semantics as python's AST operations
    # it would be good to have a reason for the inconsistency other than
    # laziness or stupidity
    def _compile_ast(self, parse_root, classname):
        # copy if we are debugging
        if self.debug_flags:
            parse_root = copy.deepcopy(parse_root)

        self._analyzed_tree = analyzer.SemanticAnalyzer(
            classname, parse_root, self.analyzer_options, self).get_ast()

        # keep a copy of the tree for debugging purposes
        if self.debug_flags:
            self._optimized_tree = copy.deepcopy(self._analyzed_tree)
        else:
            self._optimized_tree = self._analyzed_tree

        optimizer.OptimizationAnalyzer(
            self._optimized_tree, self.analyzer_options, self).optimize_ast()

        if self.debug_flags:
            self._hoisted_tree = copy.deepcopy(self._optimized_tree)
        else:
            self._hoisted_tree = self._optimized_tree

        optimizer.FinalPassAnalyzer(self._hoisted_tree, self.analyzer_options,
                                    self).optimize_ast()

        self._source_code = codegen.CodeGenerator(
            self._hoisted_tree, self, self.analyzer_options).get_code()
        return self._source_code

    def _reset(self):
        self._parse_tree = None
        self._analyzed_tree = None
        self._optimized_tree = None
        self._hoisted_tree = None
        self._source_code = None
        self.macro_context = {}

    def calculate_line_and_column(self, pos):
        if not self.src_text:
            return (0, 0)
        lineno = 1 + self.src_text.count('\n', 0, pos)
        colno = pos - (self.src_text.rfind('\n', 0, pos) + 2)
        return (lineno, colno)

    def print_stderr_message(self,
                             message,
                             pos=None,
                             is_error=False,
                             is_warning=False):
        if is_warning:
            # Print out WARNING in magenta.
            sys.stderr.write('\033[1;35mWARNING:\033[1;m ')
        elif is_error:
            # Print out ERROR in red.
            sys.stderr.write('\033[1;31mERROR:\033[1;m ')
        else:
            # Print out INFO in yellow.
            sys.stderr.write('\033[1;32mINFO:\033[1;m ')
        if pos is None:
            sys.stderr.write('%s ' % self.src_filename)
        else:
            lineno, colno = self.calculate_line_and_column(pos)
            sys.stderr.write('%s:%s:%s ' % (self.src_filename, lineno, colno))
        sys.stderr.write(message)
        sys.stderr.write('\n')

    def warn(self, message, pos=None):
        if not self.enable_warnings:
            return
        if self.warnings_as_errors:
            self.error(Warning(message), pos=pos)
        else:
            self.print_stderr_message(message, pos=pos, is_warning=True)

    def error(self, err, pos=None):
        if self.compiler_stack_traces:
            raise err
        else:
            self.print_stderr_message(str(err), pos=pos, is_error=True)
            sys.exit(1)

    def compile_template(self, src_text, classname):
        if self.tune_gc:
            gc.disable()
        try:
            self._reset()
            self._parse_tree = util.parse_template(src_text, self.xspt_mode)
            return self._compile_ast(self._parse_tree, classname)
        finally:
            if self.tune_gc:
                self._reset()
                gc.enable()
                gc.collect()

    def compile_file(self, filename):
        self.src_filename = filename
        self.classname = util.filename2classname(filename)
        self.src_text = util.read_template_file(filename)
        self.generate_line_map()
        src_code = self.compile_template(self.src_text, self.classname)
        if self.write_file:
            self.write_src_file(src_code)
        return src_code

    def generate_line_map(self):
        self.src_line_map = []
        current_line = 1
        for c in self.src_text:
            self.src_line_map.append(current_line)
            if c == '\n':
                current_line += 1

    def write_src_file(self, src_code):
        outfile_name = '%s.py' % self.classname
        relative_dir = os.path.dirname(self.src_filename)
        if self.output_directory and os.path.isabs(relative_dir):
            self.error(CompilerError(
                "can't mix output_directory and absolute paths"))

        outfile_path = os.path.join(self.output_directory, relative_dir,
                                    outfile_name)
        outfile = open(outfile_path, 'w')
        outfile.write(src_code)
        outfile.close()

    def registry_contains(self, fname):
        """Returns True if the registry contains a function."""
        return fname in self.function_name_registry

    def get_registry_value(self, fname, key):
        """Get the value associated with a key in the function registry
        for a function. The function returns None, if the function is not
        found in the registry.

        This function assumes that all registry entries are booleans.
        """
        if self.registry_contains(fname):
            decorators = self.function_name_registry[fname][-1]
            if self.new_registry_format:
                return key in decorators
            else:
                return getattr(decorators, key, False)
        return None

    # macros could be handy - and they are complex enough that they should be
    # put somewhere else. this registry allows them to be implemented just about
    # anywhere.
    #
    # macro functions look like: def macro_handler(macro_node, arg_map)
    # arg_map is a dictionary of names to values specified as parameters to the
    # macro by the template source, limited to literal values right now.
    #
    # The parse_rule argument should be string that is the name of the
    # rule you wish to parse. These rules are defined in parser.g.
    def register_macro(self, name, function, parse_rule=None):
        self.macro_registry[name] = (function, parse_rule)



# Copyright 2015 The Spitfire Authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import sys
import unittest

from spitfire.runtime import filters
from spitfire.runtime import template
from spitfire.runtime import baked
try:
    from spitfire.runtime import _template  # pylint: disable=g-import-not-at-top
except ImportError:
    _template = None


@filters.skip_filter
def is_skip():
    pass


def no_skip():
    pass

# Do not inherit from unittest.TestCase to ensure that these tests don't run.
# Add tests here and they will be run for the C and Python implementations. This
# should make sure that both implementations are equivalent.
class _TestTemplate(object):

    template_cls = None

    def setUp(self):
        self.template = self.template_cls()
        self.template._filter_function = lambda v: 'FILTERED'

    def test_skip_filter(self):
        self.assertEqual(self.template.filter_function('foo', is_skip), 'foo')

    def test_skip_filter_builtin(self):
        # Ensure skip_filter works for built-in functions as well,
        # choice of getdefaultencoding is abitrary - it's a built-in function
        is_skip_builtin = filters.skip_filter(sys.getdefaultencoding)
        self.assertEqual(self.template.filter_function('foo', is_skip_builtin),
                         'foo')
        # Ensure wrapped function works as expected
        self.assertEqual(is_skip_builtin(), sys.getdefaultencoding())

    def test_no_skip(self):
        self.assertEqual(
            self.template.filter_function('foo', no_skip), 'FILTERED')

    def test_str(self):
        self.assertEqual(self.template.filter_function('foo'), 'FILTERED')

    def test_sanitized(self):
        got = self.template.filter_function(baked.SanitizedPlaceholder('foo'))
        want = baked.SanitizedPlaceholder('foo')
        self.assertEqual(got, want)
        self.assertEqual(type(got), type(want))


class TestTemplatePy(_TestTemplate, unittest.TestCase):
    template_cls = template.get_spitfire_template_class(
        prefer_c_extension=False)


if _template is not None:

    class TestTemplateC(_TestTemplate, unittest.TestCase):
        template_cls = template.get_spitfire_template_class(
            prefer_c_extension=True)


if __name__ == '__main__':
    unittest.main()



# Copyright 2008 The Spitfire Authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.


class TreeWalkError(Exception):
    pass


def print_tree(root):
    print(TreeVisitor(root).get_text())


# perform an in-order traversal of the AST and call the generate methods
class TreeVisitor(object):

    def __init__(self, root):
        self.root = root

    def walk(self, node=None):
        if node is None:
            node = self.root
        method_name = 'visitAST%s' % node.__class__.__name__
        getattr(self, method_name, self.visitDefault)(node)

    def visitDefault(self, node):
        pass

    def visitASTTemplateNode(self, node):
        self.visitDefault(node)
        if node.import_nodes:
            for n in node.import_nodes:
                self.walk(n)
        if node.from_nodes:
            for n in node.from_nodes:
                self.walk(n)
        if node.extends_nodes:
            for n in node.extends_nodes:
                self.walk(n)
        for n in node.attr_nodes:
            self.walk(n)

        for n in node.child_nodes:
            self.walk(n)

        # if we aren't extending a template, build out the main function
        self.walk(node.main_function)

    def visitASTExtendsNode(self, node):
        self.visitDefault(node)
        for n in node.module_name_list:
            self.walk(n)

    visitASTImportNode = visitASTExtendsNode
    visitASTFromNode = visitASTExtendsNode
    visitASTAbsoluteExtendsNode = visitASTExtendsNode

    def visitASTCallFunctionNode(self, node):
        self.visitDefault(node)
        self.walk(node.expression)
        if node.arg_list:
            self.walk(node.arg_list)

    def visitASTForNode(self, node):
        self.visitDefault(node)
        self.walk(node.target_list)
        self.walk(node.expression_list)
        for n in node.child_nodes:
            self.walk(n)

    def visitASTIfNode(self, node):
        self.visitDefault(node)
        self.walk(node.test_expression)
        for n in node.child_nodes:
            self.walk(n)
        if node.else_.child_nodes:
            self.visitDefault(node.else_)
            for n in node.else_.child_nodes:
                self.walk(n)

    def visitASTAttributeNode(self, node):
        self.visitDefault(node)
        self.walk(node.default)

    def visitASTParameterListNode(self, node):
        self.visitDefault(node)
        for n in node.child_nodes:
            self.walk(n)

    visitASTArgListNode = visitASTParameterListNode
    visitASTTargetListNode = visitASTParameterListNode
    visitASTExpressionListNode = visitASTParameterListNode
    visitASTListLiteralNode = visitASTParameterListNode
    visitASTTupleLiteralNode = visitASTParameterListNode

    def visitASTDictLiteralNode(self, node):
        self.visitDefault(node)
        for key_expression, value_expression in node.child_nodes:
            self.walk(key_expression)
            self.walk(value_expression)

    def visitASTParameterNode(self, node):
        self.visitDefault(node)
        if node.default:
            self.walk(node.default)

    def visitASTGetUDNNode(self, node):
        self.visitDefault(node)
        self.walk(node.expression)

    visitASTGetAttrNode = visitASTGetUDNNode
    visitASTReturnNode = visitASTGetUDNNode
    visitASTPlaceholderSubstitutionNode = visitASTGetUDNNode
    visitASTBufferWrite = visitASTGetUDNNode
    visitASTFilterNode = visitASTGetUDNNode
    visitASTUnaryOpNode = visitASTGetUDNNode
    visitASTDoNode = visitASTGetUDNNode

    def visitASTSliceNode(self, node):
        self.visitDefault(node)
        self.walk(node.expression)
        self.walk(node.slice_expression)

    def visitASTBinOpExpressionNode(self, node):
        self.visitDefault(node)
        self.walk(node.left)
        self.walk(node.right)

    visitASTBinOpNode = visitASTBinOpExpressionNode
    visitASTAssignNode = visitASTBinOpNode

    def visitASTFunctionNode(self, node):
        self.visitDefault(node)
        if node.parameter_list:
            self.walk(node.parameter_list)

        for n in node.child_nodes:
            self.walk(n)

    visitASTDefNode = visitASTFunctionNode

    def visitASTFragmentNode(self, node):
        self.visitDefault(node)
        for n in node.child_nodes:
            self.walk(n)

    visitASTStripLinesNode = visitASTFragmentNode

    def visitASTLiteralNode(self, node):
        self.visitDefault(node)

    visitASTTextNode = visitASTLiteralNode
    visitASTWhitespaceNode = visitASTLiteralNode
    visitASTOptionalWhitespaceNode = visitASTLiteralNode


class _Finder(TreeVisitor):
    """Find the first node in the tree that satisfies a predicate."""

    def __init__(self, root, pred):
        TreeVisitor.__init__(self, root)
        self.pred = pred
        self.node_list = []

    def visitDefault(self, node):
        if self.pred(node):
            self.node_list.append(node)


def find_node(node, pred):
    """Find the first node in the tree that satisfies a predicate."""
    finder = _Finder(node, pred)
    finder.walk()
    return finder.node_list[0] if finder.node_list else None


# flatten a tree into an in-order list
class _ClearCutter(TreeVisitor):

    def __init__(self, *pargs):
        TreeVisitor.__init__(self, *pargs)
        self.node_list = []

    def visitDefault(self, node):
        self.node_list.append(node)


def flatten_tree(node):
    """Return a list of nodes in the tree."""
    cc = _ClearCutter(node)
    cc.walk()
    return cc.node_list
